
__author__    = "Andre Merzky, Ashley Z, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


'''
SLURM job adaptor implementation
'''

import re
import os
import math
import time
import datetime
import tempfile

import radical.utils as ru

from ...job           import constants   as c
from ...utils         import pty_shell   as rsups
from ...              import job         as api_job
from ...              import exceptions  as rse
from ...              import filesystem  as sfs
from ..               import base        as a_base
from ..cpi            import job         as cpi_job
from ..cpi            import decorators  as cpi_decs

from ...utils.job     import TransferDirectives

SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


# ------------------------------------------------------------------------------
# some private defs
#
_PTY_TIMEOUT = 2.0

# ------------------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.slurm_job"
_ADAPTOR_SCHEMAS       = ["slurm", "slurm+ssh", "slurm+gsissh"]
_ADAPTOR_OPTIONS       = []

# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
# TODO: FILL ALL IN FOR SLURM
_ADAPTOR_CAPABILITIES  = {
    "jdes_attributes"  : [c.NAME,
                          c.EXECUTABLE,
                          c.PRE_EXEC,
                          c.POST_EXEC,
                          c.ARGUMENTS,
                          c.ENVIRONMENT,
                          c.SPMD_VARIATION,
                          c.TOTAL_CPU_COUNT,
                          c.TOTAL_GPU_COUNT,
                          c.NUMBER_OF_PROCESSES,
                          c.PROCESSES_PER_HOST,
                          c.THREADS_PER_PROCESS,
                          c.WORKING_DIRECTORY,
                        # c.INTERACTIVE,
                          c.INPUT,
                          c.OUTPUT,
                          c.ERROR,
                          c.FILE_TRANSFER,
                          c.CLEANUP,
                          c.WALL_TIME_LIMIT,
                          c.TOTAL_PHYSICAL_MEMORY,
                          c.SYSTEM_ARCHITECTURE,
                        # c.OPERATING_SYSTEM_TYPE,
                          c.CANDIDATE_HOSTS,
                          c.QUEUE,
                          c.PROJECT,
                          c.JOB_CONTACT],
    "job_attributes"   : [c.EXIT_CODE,
                          c.EXECUTION_HOSTS,
                          c.CREATED,
                          c.STARTED,
                          c.FINISHED],
    "metrics"          : [c.STATE,
                          c.STATE_DETAIL],
    "contexts"         : {"ssh"      : "public/private keypair",
                          "x509"     : "X509 proxy for gsissh",
                          "userpass" : "username/password pair for simple ssh"}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#

        # General Notes
        # *************

        # On Stampede, returning a non-zero exit code results in the scheduler
        # putting the job into a FAILED state and assigning it an exit code of
        # 127.

        # **Example:**

        #   js = rs.job.Service("slurm+ssh://stampede")
        #   jd.executable  = '/bin/exit'
        #   jd.arguments   = ['3']
        #   job = js.create_job(jd)
        #   job.run()

        # Will return something similar to (personal account information
        # removed)::

        #   (ve) ashleyz@login1:~$ scontrol show job 309684
        #   JobId=309684 Name=SlurmJob
        #      UserId=_____ GroupId=__________
        #      Priority=3000 Account=_____ QOS=normal
        #      JobState=FAILED Reason=NonZeroExitCode Dependency=(null)
        #      Requeue=0 Restarts=0 BatchFlag=1 ExitCode=127:0
        #      RunTime=00:00:05 TimeLimit=00:01:00 TimeMin=N/A
        #      SubmitTime=2013-02-22T20:26:50 EligibleTime=2013-02-22T20:26:50
        #      StartTime=2013-02-22T20:26:50 EndTime=2013-02-22T20:26:55
        #      PreemptTime=None SuspendTime=None SecsPreSuspend=0
        #      Partition=development AllocNode:Sid=login1:12070
        #      ReqNodeList=(null) ExcNodeList=(null)
        #      NodeList=c557-401
        #      BatchHost=c557-401
        #      NumNodes=1 NumCPUs=16 CPUs/Task=1 ReqS:C:T=*:*:*
        #      MinCPUsNode=1 MinMemoryNode=0 MinTmpDiskNode=0
        #      Features=(null) Gres=(null) Reservation=(null)
        #      Shared=0 Contiguous=0 Licenses=(null) Network=(null)
        #      Command=/home1/01414/_______/.saga/adaptors/slurm_job/wrapper.sh
        #      WorkDir=/home1/01414/_______/

        # I'm not sure how to fix this for the time being.

        # Suspend/resume do not appear to be supported for regular
        # users on Stampede.

        # run_job is not supported, as there are many attributed (queues,
        # projects, etc) which need to be passed to the adaptor.  I could
        # add URL parsing so that you could pile on queue/project/jobname
        # information if this has any strong usecases, but I avoided doing
        # so for now to decrease complexity/possible confusion.

        # Cancelling a job with scontrol, puts it into a COMPLETING state, which
        # is parsed by the SLURM status parser as rs.job.RUNNING (see the
        # SLURM docs, COMPLETING is a state a job goes into when it is done
        # running but still flushing IO/etc).  Anyhow, I put some code in to
        # manually put the job into CANCELED state when the job is canceled,
        # but I'm not sure that this is reported correctly everywhere yet.

        # What exit code should be returned for a CANCELED job?


_ADAPTOR_DOC           = {
    "name"             : _ADAPTOR_NAME,
    "cfg_options"      : _ADAPTOR_OPTIONS,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "description"      : '''
        The SLURM adaptor allows to run and manage jobs on a
        `SLURM <https://computing.llnl.gov/linux/slurm/>`_ HPC cluster.

        Implementation Notes
        ********************

         - If scontrol can't find an exit code, it returns None
           (see _job_get_exit_code)
         - If scancel can't cancel a job, we raise an exception
           (see _job_cancel)
         - If we can't suspend a job with scontrol suspend, we raise an
           exception (see _job_suspend).  scontrol suspend NOT supported on
           Stampede.
         - I started to implement a dictionary to keep track of jobs locally.
           It works to the point where the unit tests are passed, but I have
           not gone over theis extensively...
         - Relating to the above, _job_get_info is written, but unused/untested
           (mostly from PBS adaptor)

        ''',
    "example": "examples/jobs/slurmjob.py",
    "schemas": {"slurm":        "connect to a local cluster",
                "slurm+ssh":    "conenct to a remote cluster via SSH",
                "slurm+gsissh": "connect to a remote cluster via GSISSH"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO          = {
    "name"             : _ADAPTOR_NAME,
    "version"          : "v0.2.1",
    "schemas"          : _ADAPTOR_SCHEMAS,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "cpis"             : [
        {
            "type"     : "radical.saga.job.Service",
            "class"    : "SLURMJobService"
        },
        {
            "type"     : "radical.saga.job.Job",
            "class"    : "SLURMJob"
        }
    ]
}


################################################################################
#
# The adaptor class
#
class Adaptor(a_base.Base):
    '''
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self):

        a_base.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        self.id_re = re.compile(r'^\[(.*)\]-\[(.*?)\]$')
        self.epoch = datetime.datetime(1970, 1, 1)


    # --------------------------------------------------------------------------
    #
    def sanity_check(self):
        pass


    def parse_id(self, id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(id)

        if  not match or len(match.groups()) != 2:
            raise rse.BadParameter("Cannot parse job id '%s'" % id)

        return(match.group(1), match.group(2))


###############################################################################
#
class SLURMJobService(cpi_job.Service):
    ''' Implements cpi_job.Service '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(SLURMJobService, self)
        _cpi_base.__init__(api, adaptor)

        # TODO make sure this formats properly and works right!
        self.exit_code_re            = re.compile(r"\bExitCode  \b=(\d*)", re.VERBOSE)
        self.scontrol_jobstate_re    = re.compile(r"\bJobState  \b=(\S*)", re.VERBOSE)
        self.scontrol_job_name_re    = re.compile(r"\bJobName   \b=(\S*)", re.VERBOSE)
        self.scontrol_create_time_re = re.compile(r"\bSubmitTime\b=(\S*)", re.VERBOSE)
        self.scontrol_start_time_re  = re.compile(r"\bStartTime \b=(\S*)", re.VERBOSE)
        self.scontrol_end_time_re    = re.compile(r"\bEndTime   \b=(\S*)", re.VERBOSE)
        self.scontrol_comp_time_re   = re.compile(r"\bRunTime   \b=(\S*)", re.VERBOSE)
        self.scontrol_exec_hosts_re  = re.compile(r"\bNodeList  \b=(\S*)", re.VERBOSE)

        # these are the commands that we need in order to interact with SLURM
        # the adaptor will try to find them when it first opens the shell
        # connection, and bails out in case they are not available.
        self._commands = {'sbatch'  : None,
                          'squeue'  : None,
                          'scontrol': None,
                          'scancel' : None}


    # --------------------------------------------------------------------------
    #
    def __del__(self):

        try:
            if self.shell:
                del(self.shell)
        except:
            pass


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        '''
        Service instance constructor
        '''

        self.rm      = rm_url
        self.session = session

        self.jobs = {}
        self._open()

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def close(self):

        if self.shell:
            self.shell.finalize(True)


    # --------------------------------------------------------------------------
    #
    def _open(self):
        '''
        Open our persistent shell for this job adaptor.  We use
        the pty_shell functionality for this.
        '''

        # check to see what kind of connection we will want to create
        if   self.rm.schema == "slurm":        shell_schema = "fork://"
        elif self.rm.schema == "slurm+ssh":    shell_schema = "ssh://"
        elif self.rm.schema == "slurm+gsissh": shell_schema = "gsissh://"
        else: raise rse.IncorrectURL("Schema %s not supported by SLURM adaptor."
                                     % self.rm.schema)

        # <scheme>://<user>:<pass>@<host>:<port>/<path>?<query>#<fragment>
        # build our shell URL
        shell_url = shell_schema

        # did we provide a username and password?
        if self.rm.username and self.rm.password:
            shell_url += self.rm.username + ":" + self.rm.password + "@"

        # only provided a username
        if self.rm.username and not self.rm.password:
            shell_url += self.rm.username + "@"

        # add hostname
        shell_url += self.rm.host

        # add port
        if  self.rm.port:
            shell_url += ":" + str(self.rm.port)

        shell_url = ru.Url(shell_url)

        # establish shell connection
        self._logger.debug("Opening shell of type: %s" % shell_url)
        self.shell = rsups.PTYShell(shell_url, self.session, self._logger)

        # verify our SLURM environment contains the commands we need for this
        # adaptor to work properly
        self._logger.debug("Verifying existence of remote SLURM tools.")
        for cmd in list(self._commands.keys()):
            ret, out, _ = self.shell.run_sync("which %s " % cmd)
            if ret != 0:
                message = "Error finding SLURM tool %s on remote server %s!\n" \
                          "Locations searched:\n%s\n" \
                          "Is SLURM installed on that machine? " \
                          "If so, is your remote SLURM environment "\
                          "configured properly? " % (cmd, self.rm, out)
                raise rse.NoSuccess._log(self._logger, message)

        self._logger.debug("got cmd prompt (%s)(%s)" % (ret, out))

        self.rm.detected_username = self.rm.username
        # figure out username if it wasn't made explicit
        # important if .ssh/config info read+connected with
        # a different username than what we expect
        if not self.rm.username:
            self._logger.debug("No username provided in URL %s, so we are"
                               " going to find it with whoami" % self.rm)
            ret, out, _ = self.shell.run_sync("whoami")
            self.rm.detected_username = out.strip()
            self._logger.debug("Username detected as: %s",
                               self.rm.detected_username)

        _, out, _ = self.shell.run_sync('scontrol --version')
        self._version = out.split()[1].strip()
        self._logger.info('slurm version: %s' % self._version)

        ppn_pat   = '\'s/.*\\(CPUTot=[0-9]*\\).*/\\1/g\''
        ppn_cmd   = 'scontrol show nodes ' + \
                    '| grep CPUTot'        + \
                    '| sed -e ' + ppn_pat  + \
                    '| sort '              + \
                    '| uniq -c '           + \
                    '| cut -f 2 -d = '     + \
                    '| xargs echo'
        _, out, _ = self.shell.run_sync(ppn_cmd)
        ppn_vals  = [o.strip() for o in out.split() if o.strip()]
        if len(ppn_vals) >= 1: self._ppn = int(ppn_vals[0])
        else                 : self._ppn = None

        if 'stampede2' in self.rm.host.lower():
            # FIXME: this only works on the KNL nodes
            self._ppn = 68

        elif 'traverse' in self.rm.host.lower():
            self._ppn = 32

        elif 'frontera' in self.rm.host.lower():
            # not that this is incorrect for the rtx queue
            self._ppn = 56

        elif 'comet' in self.rm.host.lower():
            self._ppn = 24

        elif 'longhorn' in self.rm.host.lower():
            # FIXME: other option - get it later by `processes_per_host`
            self._ppn = 40

        self._logger.info("ppn: %s", self._ppn)


    # --------------------------------------------------------------------------
    #
    def _close(self):
        '''
        Close our shell connection
        '''

        del(self.shell)
        self.shell = None


    # --------------------------------------------------------------------------
    #
    def _handle_file_transfers(self, ft, mode):
        """
        if mode == 'in' : perform sanity checks on all staging directives.

        if mode == 'in' : stage files to   condor submission site
        if mode == 'out': stage files from condor submission site
        """

        td = TransferDirectives(ft)

        assert(mode in ['in', 'out'])

        if mode == 'in':

            if td.in_append:
                raise Exception('File append (>>) not supported')

            if td.out_append:
                raise Exception('File append (<<) not supported')

            if td.in_overwrite:

                for (local, remote) in td.in_overwrite:

                    source = local
                    target = remote

                    self._logger.info("Transferring in %s to %s", source, target)
                    self.shell.stage_to_remote(source, target)

        elif mode == 'out':

            if td.out_overwrite:

                for (local, remote) in td.out_overwrite:

                    source = remote
                    target = local

                    self._logger.info("Transferring out %s to %s", source, target)
                    self.shell.stage_from_remote(source, target)


    # --------------------------------------------------------------------------
    #
    #
    def _job_run(self, jd):
        '''
        runs a job on the wrapper via pty, and returns the job id
        '''

        # define a bunch of default args
        exe                 = jd.executable
        pre                 = jd.as_dict().get(c.PRE_EXEC)
        post                = jd.as_dict().get(c.POST_EXEC)
        args                = jd.as_dict().get(c.ARGUMENTS, [])
        env                 = jd.as_dict().get(c.ENVIRONMENT, dict())
        cwd                 = jd.as_dict().get(c.WORKING_DIRECTORY)
        job_name            = jd.as_dict().get(c.NAME)
        spmd_variation      = jd.as_dict().get(c.SPMD_VARIATION)
        cpu_count           = jd.as_dict().get(c.TOTAL_CPU_COUNT)
        gpu_count           = jd.as_dict().get(c.TOTAL_GPU_COUNT)
        n_procs             = jd.as_dict().get(c.NUMBER_OF_PROCESSES)
        processes_per_host  = jd.as_dict().get(c.PROCESSES_PER_HOST)
        output              = jd.as_dict().get(c.OUTPUT, "radical.saga.stdout")
        error               = jd.as_dict().get(c.ERROR,  "radical.saga.stderr")
        file_transfer       = jd.as_dict().get(c.FILE_TRANSFER)
        wall_time           = jd.as_dict().get(c.WALL_TIME_LIMIT)
        queue               = jd.as_dict().get(c.QUEUE)
        project             = jd.as_dict().get(c.PROJECT)
        total_memory        = jd.as_dict().get(c.TOTAL_PHYSICAL_MEMORY)
        sys_arch            = jd.as_dict().get(c.SYSTEM_ARCHITECTURE)
        job_contact         = jd.as_dict().get(c.JOB_CONTACT)
        c_hosts             = jd.as_dict().get(c.CANDIDATE_HOSTS)

        cpu_arch            = sys_arch.get('cpu')
        gpu_arch            = sys_arch.get('gpu')

        # check to see what's available in our job description
        # to override defaults

        # try to create the working directory (if defined)
        # NOTE: this assumes a shared filesystem between login node and
        #       compute nodes.
        if cwd:

            self._logger.info("Creating working directory %s" % cwd)
            ret, out, _ = self.shell.run_sync("mkdir -p %s"   % cwd)

            if ret:
                raise rse.NoSuccess("Couldn't create workdir: %s" % out)

        self._handle_file_transfers(file_transfer, mode='in')

        if isinstance(c_hosts, list):
            c_hosts = ','.join(c_hosts)

        if isinstance(job_contact, list):
            job_contact = job_contact[0]

        if project and ':' in project:
            account, reservation = project.split(':', 1)
        else:
            account, reservation = project, None

        script = "#!/bin/sh\n\n"

        # make sure we have something for cpu_count
        if not cpu_count:
            cpu_count = 1

        # make sure we have something for n_procs
        if not n_procs:
            n_procs = cpu_count

        # get memory_per_node from total_memory and make sure it is not None
        memory_per_node = total_memory or 0

        # define n_nodes and recalculate memory_per_node (if self._ppn is set)
        n_nodes = None
        if self._ppn:

            # exception(s) for earlier defined `self._ppn`
            if 'frontera' in self.rm.host.lower() and \
                    queue and 'rtx' in queue.lower():
                self._ppn = 16  # other option is to use: processes_per_host

            n_nodes = int(math.ceil(float(cpu_count) / self._ppn))
            memory_per_node = int(memory_per_node / float(n_nodes))

        elif total_memory:
            raise rse.NotImplemented(
                'cannot allocate memory, node number unknown')

        if spmd_variation:
            if spmd_variation.lower() not in 'mpi':
                raise rse.BadParameter("Slurm cannot handle spmd variation '%s'"
                                        % spmd_variation)
            mpi_cmd = 'mpirun -n %d ' % n_procs

        else:
            # we start N independent processes
            mpi_cmd = ''

            if  'stampede2' in self.rm.host.lower() or \
                'longhorn'  in self.rm.host.lower():

                assert(n_nodes), 'need unique number of cores per node'
                script += "#SBATCH -N %d\n" % n_nodes
                script += "#SBATCH -n %s\n" % n_procs

            elif 'frontera'  in self.rm.host.lower() or \
                 'rhea'      in self.rm.host.lower():

                assert(n_nodes), 'need unique number of cores per node'
                script += "#SBATCH -N %d\n" % n_nodes

            elif self._version in ['17.11.5', '18.08.0', '18.08.3']:

                assert(n_nodes), 'need unique number of cores per node'
                script += "#SBATCH -N %d\n" % n_nodes
                script += "#SBATCH --ntasks=%s\n" % n_procs

            else:
                script += "#SBATCH --ntasks=%s\n" % n_procs

            if not processes_per_host:
                script += "#SBATCH --cpus-per-task=%s\n" \
                        % (int(cpu_count / n_procs))

            else:
                script += "#SBATCH --ntasks-per-node=%s\n" % processes_per_host

        # target host specifica
        # FIXME: these should be moved into resource config files
        self._logger.debug ("submit SLURM script to %s", self.rm)
        if 'bridges' in self.rm.host.lower():

            if gpu_count:
                # gres resources are specified *per node*
                assert(n_nodes), 'need unique number of cores per node'

                if gpu_arch: gpu_arch = gpu_arch.lower()
                else       : gpu_arch = 'p100'
                if gpu_arch == 'k80': count = 4
                else                : count = 2
                # Make sure we take a full GPU node
                script += "#SBATCH --gres=gpu:%s:%d\n" % (gpu_arch, count)

            # use '-C EGRESS' to enable outbound network
            script += "#SBATCH -C EGRESS\n"

        elif 'comet' in self.rm.host.lower():

            if gpu_count:
                # gres resources are specified *per node*
                assert(n_nodes), 'need unique number of cores per node'

                # if no `gpu_arch` then first available gpu node (either type)
                # gpu types are "p100" and "k80"
                if gpu_arch: gpu_arch_str = ':%s' % gpu_arch.lower()
                else       : gpu_arch_str = ''
                count = 4
                # Make sure we take a full GPU node
                script += "#SBATCH --gres=gpu%s:%d\n" % (gpu_arch_str, count)

        elif 'tiger' in self.rm.host.lower():

            if gpu_count:
                # gres resources are specified *per node*
                assert(n_nodes), 'need unique number of cores per node'
                count = int(gpu_count / n_nodes)

                if count:
                    script += "#SBATCH --gres=gpu:%s\n" % count

        elif 'cori' in self.rm.host.lower():

            # Set to "haswell" for Haswell nodes, to "knl,quad,cache" (or other
            # modes) for KNL, etc.
            if cpu_arch : script += "#SBATCH -C %s\n"     % cpu_arch
            if gpu_count: script += "#SBATCH --gpus=%s\n" % gpu_count

        elif queue == 'tmp3':

            # this is a special queue, which is associated with SuperMUC-NG,
            # but since there is no machine name in config data we only track
            # this queue name to set SLURM QoS option
            script += "#SBATCH --qos=nolimit\n"
            self._logger.debug("SLURM QoS is set (SuperMUC-NG only)\n")

        else:

            if gpu_count: script += "#SBATCH --gpus=%s\n" % gpu_count

        if cwd:
            if 'frontera' in self.rm.host.lower() or \
               'longhorn' in self.rm.host.lower() or \
               'tiger'    in self.rm.host.lower() or \
               'traverse' in self.rm.host.lower() or \
               'rhea'     in self.rm.host.lower():
                script += "#SBATCH --chdir %s\n"   % cwd
            else:
                script += "#SBATCH --workdir %s\n" % cwd

        if output         : script += "#SBATCH --output %s\n"      % output
        if error          : script += "#SBATCH --error %s\n"       % error
        if queue          : script += "#SBATCH --partition %s\n"   % queue
        if job_name       : script += '#SBATCH -J "%s"\n'          % job_name
        if c_hosts        : script += "#SBATCH --nodelist=%s\n"    % c_hosts
        if job_contact    : script += "#SBATCH --mail-user=%s\n"   % job_contact
        if account        : script += "#SBATCH --account %s\n"     % account
        if reservation    : script += "#SBATCH --reservation %s\n" % reservation
        if wall_time      : script += "#SBATCH --time %02d:%02d:00\n" \
                                         % (int(wall_time / 60), wall_time % 60)
        if memory_per_node: script += "#SBATCH --mem=%s\n" % memory_per_node

        if env:
            script += "\n## ENVIRONMENT\n"
            for key,val in env.items():
                script += 'export "%s"="%s"\n'  %  (key, val)

        if pre:
            script += "\n## PRE_EXEC\n" + "\n".join(pre)
            script += '\n'

        # create our commandline
        script += "\n## EXEC\n"
        script += '%s%s %s' % (mpi_cmd, exe, ' '.join(args))
        script += '\n'

        if post:
            script += "\n## POST_EXEC\n" + '\n'.join(post)
            script += '\n'

        # write script into a tmp file for staging
        self._logger.info("SLURM script generated:\n%s" % script)

        tgt = os.path.basename(tempfile.mktemp(suffix='.slurm', prefix='tmp_'))
        self.shell.write_to_remote(src=script, tgt=tgt)

        # submit the job
        ret, out, _ = self.shell.run_sync("sbatch '%s'; echo rm -f '%s'" % (tgt,tgt))

        self._logger.debug("submit SLURM script (%s) (%s)" % (tgt, ret))

        # find out what our job ID is
        # TODO: Could make this more efficient
        job_id = None
        for line in out.split("\n"):
            if "Submitted batch job" in line:
                job_id = "[%s]-[%s]" % (self.rm, int(line.split()[-1:][0]))
                break

        # if we have no job ID, there's a failure...
        if not job_id:
            raise rse.NoSuccess._log(self._logger,
                             "Couldn't get job id from submitted job!"
                              " sbatch output:\n%s" % out)

        self._logger.debug("started job %s" % job_id)
        self._logger.debug("Batch system output:\n%s" % out)

        # create local jobs dictionary entry
        self.jobs[job_id] = {'state'      : c.PENDING,
                             'create_time': None,
                             'start_time' : None,
                             'end_time'   : None,
                             'comp_time'  : None,
                             'exec_hosts' : None,
                             'gone'       : False,
                             'output'     : output,
                             'error'      : error,
                             'stdout'     : None,
                             'stderr'     : None,
                             'ft'         : file_transfer,
                             }
        return job_id


    # --------------------------------------------------------------------------
    #
    # FROM STAMPEDE'S SQUEUE MAN PAGE
    #
    # JOB STATE CODES
    #    Jobs typically pass through several states in the course of their
    #    execution.  The typical states are PENDING, RUNNING, SUSPENDED,
    #    COMPLETING, and COMPLETED.   An  explanation  of  each state follows.
    #
    #    CA  CANCELED        Job was explicitly cancelled by the user or system
    #                        administrator.  The job may or may not have been
    #                        initiated.
    #    CD  COMPLETED       Job has terminated all processes on all nodes.
    #    CF  CONFIGURING     Job has been allocated resources, but are waiting
    #                        for them to become ready for use (e.g. booting).
    #    CG  COMPLETING      Job is in the process of completing. Some processes
    #                        on some nodes may still be active.
    #    F   FAILED          Job terminated with non-zero exit code or other
    #                        failure condition.
    #    NF  NODE_FAIL       Job terminated due to failure of one or more
    #                        allocated nodes.
    #    PD  PENDING         Job is awaiting resource allocation.
    #    PR  PREEMPTED       Job terminated due to preemption.
    #    R   RUNNING         Job currently has an allocation.
    #    S   SUSPENDED       Job has an allocation, but execution has been
    #                        suspended.
    #    TO  TIMEOUT         Job terminated upon reaching its time limit.
    #
    def _slurm_to_saga_state(self, slurmjs):
        '''
        translates a slurm one-letter state to saga
        '''

        if   slurmjs in ['CA', "CANCELLED"  ]: return c.CANCELED
        elif slurmjs in ['CD', "COMPLETED"  ]: return c.DONE
        elif slurmjs in ['CF', "CONFIGURING"]: return c.PENDING
        elif slurmjs in ['CG', "COMPLETING" ]: return c.RUNNING
        elif slurmjs in ['F' , "FAILED"     ]: return c.FAILED
        elif slurmjs in ['NF', "NODE_FAIL"  ]: return c.FAILED
        elif slurmjs in ['PD', "PENDING"    ]: return c.PENDING
        elif slurmjs in ['PR', "PREEMPTED"  ]: return c.CANCELED
        elif slurmjs in ['R' , "RUNNING"    ]: return c.RUNNING
        elif slurmjs in ['S' , "SUSPENDED"  ]: return c.SUSPENDED
        elif slurmjs in ['TO', "TIMEOUT"    ]: return c.CANCELED
        else                                 : return c.UNKNOWN


    # --------------------------------------------------------------------------
    #
    def _job_cancel(self, job):
        '''
        Given a job id, attempt to cancel it through use of commandline
        scancel.  Raises exception when unsuccessful.
        '''

        if job._state in c.FINAL:
            # job is already final - nothing to do
            return

        if job._state in [c.NEW]:
            # job is not yet submitted - nothing to do
            job._state = c.CANCELED

        if not job._id:
            # uh oh - what to do?
            raise rse.NoSuccess._log(self._logger,
                    "Could not cancel job: no job ID")

        rm,  pid    = self._adaptor.parse_id(job._id)
        ret, out, _ = self.shell.run_sync("scancel %s" % pid)

        if ret != 0:
            raise rse.NoSuccess._log(self._logger,
                    "Could not cancel job %s because: %s" % (pid, out))

        job._state = c.CANCELED


    # --------------------------------------------------------------------------
    #
    def _job_suspend(self, job):
        '''
        Attempt to suspend a job with commandline scontrol.  Raise
        exception when unsuccessful.
        '''

        if job._state in [c.DONE, c.FAILED, c.CANCELED, c.NEW, c.SUSPENDED]:
            raise rse.IncorrectState._log(self._logger,
                    "Could not suspend job %s [%s]" % (job._id, job._state))


        rm,  pid    = self._adaptor.parse_id(job._id)
        ret, out, _ = self.shell.run_sync("scontrol suspend %s" % pid)

        if ret == 0:
            return True

        # check to see if the error was a permission error
        elif "Access/permission denied" in out:
            raise rse.PermissionDenied._log(self._logger,
                    "Could not suspend job %s because: %s" % (pid, out))

        # it's some other error
        else:
            raise rse.NoSuccess._log(self._logger,
                    "Could not suspend job %s because: %s" % (pid, out))


    # --------------------------------------------------------------------------
    #
    def _job_resume(self, job):
        '''
        Attempt to resume a job with commandline scontrol.  Raise
        exception when unsuccessful.
        '''

        if job._state in [c.DONE, c.FAILED, c.CANCELED, c.NEW, c.RUNNING]:
            raise rse.IncorrectState._log(self._logger,
                    "Could not resume job %s [%s]" % (job._id, job._state))


        rm,  pid    = self._adaptor.parse_id(job._id)
        ret, out, _ = self.shell.run_sync("scontrol resume %s" % pid)

        if ret == 0:
            return True

        # check to see if the error was a permission error
        elif "Access/permission denied" in out:
            raise rse.PermissionDenied._log(self._logger,
                    "Could not suspend job %s because: %s" % (pid, out))

        # it's some other error
        else:
            raise rse.NoSuccess._log(self._logger,
                    "Could not resume job %s because: %s" % (pid, out))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        '''
        Implements cpi_job.Service.create_job()
        '''

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service"    : self,
                         "job_description": jd,
                         "job_schema"     : self.rm.schema,
                         "reconnect"      : False}

        return api_job.Job(_adaptor=self._adaptor, _adaptor_state=adaptor_state)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        '''
        Implements cpi_job.Service.get_url()
        '''
        return self.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        '''
        Implements rs.adaptors.cpi.job.Service.list()
        '''

        # ashleyz@login1:~$ squeue -h -o "%i" -u ashleyz
        # 255042
        # 255035
        # 255028
        # 255018

        # this line gives us a nothing but jobids for our user
        ret, out, _ = self.shell.run_sync('squeue -h -o "%%i" -u %s'
                                          % self.rm.detected_username)

        # mangle our results into the proper id format
        output = ["[%s]-[%s]" % (self.rm, i) for i in out.strip().split("\n")]
        return output


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, jobid):

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.  The job adaptor will run 'scontrol
        # show job $jobid' to complement the information.
        adaptor_state = {"job_service"    : self,
                         "job_description": api_job.Description(),
                         "job_schema"     : self.rm.schema,
                         "reconnect"      : True,
                         "reconnect_jobid": jobid
                        }
        return api_job.Job(_adaptor=self._adaptor,
                            _adaptor_state=adaptor_state)


    # --------------------------------------------------------------------------
    #
    def container_run(self, jobs):

        for job in jobs:
            job.run()


    # --------------------------------------------------------------------------
    #
    def container_wait(self, jobs, mode, timeout):

        # TODO: this is not optimized yet
        for job in jobs:
            job.wait()


    # --------------------------------------------------------------------------
    #
    def container_cancel(self, jobs, timeout):

        # TODO: this is not optimized yet
        for job in jobs:
            job.cancel(timeout)


    # --------------------------------------------------------------------------
    #
    def container_get_states(self, jobs):

        # TODO: this is not optimized yet
        states = list()
        for job in jobs:
            states.append(job.get_state())
        return states


# ------------------------------------------------------------------------------
#
class SLURMJob(cpi_job.Job):

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(SLURMJob, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, job_info):

        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        # the js is responsible for job bulk operations -- which
        # for jobs only work for run()
        self._container       = self.js
        self._method_type     = "run"

        # initialize job attribute values
        self._id              = None
        self._name            = self.jd.as_dict().get(c.NAME, 'saga')
        self._state           = c.NEW
        self._exit_code       = None
        self._exception       = None
        self._started         = None
        self._finished        = None

        # think "reconnect" in terms of "reloading" job id, _NOT_
        # physically creating a new network connection
        if job_info['reconnect']:
            self._id   = job_info['reconnect_jobid']
            other_info = self._job_get_info()
            self._name = other_info.get('job_name')
            self._started = True
        else:
            self._started = False

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def _job_get_info(self):
        '''
        use scontrol to grab job info
        NOT CURRENTLY USED/TESTED, here for later
        '''

        # prev. info contains the info collect when _job_get_info
        # was called the last time
        prev_info = self.js.jobs.get(self._id)

        # if the 'gone' flag is set, there's no need to query the job
        # state again. it's gone forever
        if prev_info:
            if prev_info.get('gone', False):
                self._logger.debug("Job is gone.")
                return prev_info

        # curr. info will contain the new job info collect. it starts off
        # as a copy of prev_info (don't use deepcopy because there is an API
        # object in the dict -> recursion)
        curr_info = dict()

        if prev_info:
            curr_info['job_id'     ] = prev_info.get('job_id'     )
            curr_info['job_name'   ] = prev_info.get('job_name'   )
            curr_info['state'      ] = prev_info.get('state'      )
            curr_info['create_time'] = prev_info.get('create_time')
            curr_info['start_time' ] = prev_info.get('start_time' )
            curr_info['end_time'   ] = prev_info.get('end_time'   )
            curr_info['comp_time'  ] = prev_info.get('comp_time'  )
            curr_info['exec_hosts' ] = prev_info.get('exec_hosts' )
            curr_info['gone'       ] = prev_info.get('gone'       )
            curr_info['output'     ] = prev_info.get('output'     )
            curr_info['error'      ] = prev_info.get('error'      )
            curr_info['stdout'     ] = prev_info.get('stdout'     )
            curr_info['stderr'     ] = prev_info.get('stderr'     )
            curr_info['ft'         ] = prev_info.get('ft'         )
        else:
            curr_info['job_id'     ] = None
            curr_info['job_name'   ] = None
            curr_info['state'      ] = None
            curr_info['create_time'] = None
            curr_info['start_time' ] = None
            curr_info['end_time'   ] = None
            curr_info['comp_time'  ] = None
            curr_info['exec_hosts' ] = None
            curr_info['gone'       ] = None
            curr_info['output'     ] = None
            curr_info['error'      ] = None
            curr_info['stdout'     ] = None
            curr_info['stderr'     ] = None
            curr_info['ft'         ] = None

        rm, pid = self._adaptor.parse_id(self._id)

        # update current info with scontrol
        ret, out, _ = self.js.shell.run_sync('scontrol show job %s' % pid)

        # out is comprised of a set of space-limited words like this:
        #
        # ----------------------------------------------------------------------
        # $ scontrol show job 8101313
        #    JobId=8101313 JobName=pilot.0000 UserId=tg803521(803521)
        #    GroupId=G-81625(81625) Priority=1701 Nice=0 Account=TG-MCB090174
        #    QOS=normal JobState=RUNNING Reason=None Dependency=(null) Requeue=0
        #    Restarts=0 BatchFlag=1 Reboot=0 ExitCode=0:0 RunTime=00:00:25
        #    TimeLimit=00:15:00 TimeMin=N/A SubmitTime=2017-01-11T15:47:19
        #    EligibleTime=2017-01-11T15:47:19 StartTime=2017-01-11T15:47:19
        #    EndTime=2017-01-11T16:02:19 PreemptTime=None SuspendTime=None
        #    SecsPreSuspend=0 Partition=development AllocNode:Sid=login3:2886
        #    ReqNodeList=(null) ExcNodeList=(null) NodeList=c557-[901-904]
        #    BatchHost=c557-901 NumNodes=4 NumCPUs=64 CPUs/Task=1
        #    ReqB:S:C:T=0:0:*:* TRES=cpu=64,node=4 Socks/Node=*
        #    NtasksPerN:B:S:C=0:0:*:* CoreSpec=* MinCPUsNode=1 MinMemoryNode=0
        #    MinTmpDiskNode=0 Features=(null) Gres=(null) Reservation=(null)
        #    Shared=0 Contiguous=0 Licenses=(null) Network=(null)
        #    Command=/home1/01083/tg803521/tmp_egGk1n.slurm
        #    WorkDir=/work/01083/
        #    StdIn=/dev/null
        #    StdOut=/work/01083/bootstrap_1.out
        #    StdErr=/work/01083/bootstrap_1.err
        #    Power= SICP=0
        # ----------------------------------------------------------------------
        #
        # so we split on spaces and newlines, and then on '=' to get
        # key-value-pairs.

        elems = out.split()
        data  = dict()
        for elem in sorted(elems):

            parts = elem.split('=', 1)

            if len(parts) == 1:
                # default if no '=' is found
                parts.append(None)

            # ignore non-splittable ones
            key, val = parts
            if val in ['', '(null)']:
                val = None
            self._logger.info('%-20s := %s', key, val)
            data[key] = val

        if data.get('JobState'):
            curr_info['state'] = self.js._slurm_to_saga_state(data['JobState'])
        else:
            curr_info['state'] = self._job_get_state(self._id)

        # update exit code
        if data.get('ExitCode'):
            curr_info['exit_code'] = data['ExitCode'].split(':')[0]
        else:
            curr_info['exit_code'] = self._job_get_state(self._id)

        curr_info['job_name'   ] = data.get('JobName')
      # curr_info['create_time'] = data.get('SubmitTime')
      # curr_info['start_time' ] = data.get('StartTime')
      # curr_info['end_time'   ] = data.get('EndTime')
        curr_info['comp_time'  ] = data.get('RunTime')
        curr_info['exec_hosts' ] = data.get('NodeList')

        # Alas, time stamps are not in EPOCH, and do not contain time zone info,
        # so we set approximate values here
        now = time.time()
        if not curr_info['create_time']: curr_info['create_time'] = now

        if curr_info['state'] in [c.RUNNING] + c.FINAL:
            if not curr_info['start_time' ]: curr_info['start_time' ] = now

        if curr_info['state'] in c.FINAL:

            if not curr_info['end_time' ]: curr_info['end_time' ] = now

            if curr_info['stdout'] is None:

                if curr_info['output'] is None:
                    curr_info['output'] = data.get('StdOut')

                ret, out, err = self.js.shell.run_sync(
                                                 'cat %s' % curr_info['output'])
                if ret: curr_info['stdout'] = None
                else  : curr_info['stdout'] = out

            if curr_info['stderr'] is None:

                if curr_info['error'] is None:
                    curr_info['error'] = data.get('StdErr')

                ret, out, err = self.js.shell.run_sync(
                                                  'cat %s' % curr_info['error'])
                if ret: curr_info['stderr'] = None
                else  : curr_info['stderr'] = out

            self.js._handle_file_transfers(curr_info['ft'], mode='out')

            curr_info['gone'] = True

        self.js.jobs[self._id] = curr_info

        return curr_info


    # --------------------------------------------------------------------------
    #
    def _job_get_state(self, job_id):
        '''
        get the job state from the wrapper shell
        '''

        # if the state is NEW and we haven't sent out a run command, keep
        # it listed as NEW
        if self._state == c.NEW and not self._started:
            return c.NEW

        # if the state is DONE, CANCELED or FAILED, it is considered
        # final and we don't need to query the backend again
        if self._state in c.FINAL:
            return self._state

        rm, pid = self._adaptor.parse_id(job_id)

        try:
            ret, out, _ = self.js.shell.run_sync('scontrol show job %s' % pid)
            match       = self.js.scontrol_jobstate_re.search(out)

            if match:
                slurm_state = match.group(1)
            else:
                # no jobstate found from scontrol
                # the job may have finished a while back, use sacct to
                # look at the full slurm history
                slurm_state = self._sacct_jobstate_match(pid)
                if not slurm_state:
                    # no jobstate found in slurm
                    return c.UNKNOWN

            return self.js._slurm_to_saga_state(slurm_state)

        except Exception as e:
            self._logger.exception('failed to get job state')
            raise rse.NoSuccess("Error getting the job state for "
                                "job %s:\n%s" % (pid, e)) from e

        raise rse.NoSuccess._log(self._logger, "Internal SLURM adaptor error"
                                 " in _job_get_state")


    # --------------------------------------------------------------------------
    #
    def _sacct_jobstate_match(self, pid):
        '''
        get the job state from the slurm accounting data
        '''

        ret, sacct_out, _ = self.js.shell.run_sync(
            "sacct --format=JobID,State --parsable2 --noheader --jobs=%s" % pid)

        # output will look like:
        # 500723|COMPLETED
        # 500723.batch|COMPLETED
        # or:
        # 500682|CANCELLED by 900369
        # 500682.batch|CANCELLED

        try:
            for line in sacct_out.strip().split('\n'):

                slurm_id, slurm_state = line.split('|', 1)

                if slurm_id == pid and slurm_state:
                    return slurm_state.split()[0].strip()

        except Exception:
            self._logger.warning('cannot parse sacct output:\n%s' % sacct_out)

        return None


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):

        self._state = self._job_get_state(self._id)
        return self._state


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_stdout(self):

        out = self._job_get_info()['stdout']
        if out is None:
            out = ''
          # raise rse.NoSuccess("Couldn't fetch stdout (js reconnected?)")
        return out


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_stderr(self):

        err = self._job_get_info()['stderr']
        if err is None:
            err = ''
          # raise rse.NoSuccess("Couldn't fetch stderr (js reconnected?)")
        return err


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description(self):

        return self.jd


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):

        return self.js.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):

        time_start = time.time()
        rm, pid    = self._adaptor.parse_id(self._id)

        while True:

            state = self._job_get_state(self._id)
            self._logger.debug("wait() for job id %s:%s" % (self._id, state))

            if state == c.UNKNOWN:
                raise rse.IncorrectState("cannot get job state")

            if state in c.FINAL:
                self._job_get_info()
                return True

            # check if we hit timeout
            if timeout >= 0:
                if time.time() - time_start > timeout:
                    return False

            # avoid busy poll
            time.sleep(0.5)


    # --------------------------------------------------------------------------
    #
    # In general, the job ID is something which is generated by the adaptor or
    # by the backend, and the user should not interpret it.  Two caveats though:
    #
    # (a) The ID MUST remain constant once it is assigned to a job (imagine an
    # application hashes on job ids, for example.
    #
    # (b) the ID SHOULD follow the scheme [service_url]-[backend-id] -- and in
    # that case, we should make sure that the URL part of the ID can be used to
    # create a new job service instance.
    #
    @SYNC_CALL
    def get_id(self):

        return self._id


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name(self):

        if not self._name:
            self._name = self._job_get_info()['job_name']
        return self._name


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):

        # FIXME: use cache
        return self._job_get_info()['exit_code']


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def suspend(self):

        return self.js._job_suspend(self)


    # --------------------------------------------------------------------------
    @SYNC_CALL
    def resume(self):

        return self.js._job_resume(self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):

        # FIXME: use cache
        # FIXME: convert to EOPCH
        return self._job_get_info()['create_time']


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):

        # FIXME: use cache
        # FIXME: convert to EPOCH
        return self._job_get_info()['start_time']


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):

        # FIXME: use cache
        # FIXME: convert to EPOCH
        return self._job_get_info()['end_time']


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):

        # FIXME: use cache
        return self._job_get_info()['exec_hosts']


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):

        self.js._job_cancel(self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):

        self._id      = self.js._job_run(self.jd)
        self._started = True


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import os
import subprocess

from ...exceptions import *
from ..            import base
from ..cpi         import SYNC_CALL, ASYNC_CALL
from ..cpi         import context as cpi
from ...           import context as api
from ...exceptions import *


######################################################################
#
# adaptor meta data
#
_ADAPTOR_NAME          = 'adaptor_myproxy'
_ADAPTOR_SCHEMAS       = ['MyProxy']

_ADAPTOR_CAPABILITIES  = {
    'attributes'       : [api.TYPE,
                          api.SERVER,
                          api.USER_ID,
                          api.USER_PASS,
                          api.LIFE_TIME]
}

_ADAPTOR_DOC           = {
    'name'             : _ADAPTOR_NAME,
    'capabilities'     : _ADAPTOR_CAPABILITIES,
    'description'      : """This adaptor fetches an X509 proxy from
                            MyProxy when it is added to a saga.Session.""",
    'schemas'          : {'myproxy' : 'this adaptor can only interact with myproxy backends'},
}

_ADAPTOR_INFO          = {
    'name'             : _ADAPTOR_NAME,
    'version'          : 'v0.1',
    'schemas'          : _ADAPTOR_SCHEMAS,
    'cpis'             : [{
        'type'         : 'radical.saga.Context',
        'class'        : 'ContextMyProxy'
        }
    ]
}


###############################################################################
# The adaptor class

class Adaptor (base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    def __init__ (self) :

        base.Base.__init__ (self, _ADAPTOR_INFO)

        # there are no default myproxy contexts
        self._default_contexts = []
        self.base_workdir = self._cfg.get('base_workdir', os.getcwd())


    def sanity_check (self) :
        pass


    def _get_default_contexts (self) :

        return self._default_contexts



######################################################################
#
# job adaptor class
#
class ContextMyProxy (cpi.Context) :

    def __init__ (self, api, adaptor) :

        _cpi_base = super  (ContextMyProxy, self)
        _cpi_base.__init__ (api, adaptor)
        self.base_workdir = adaptor.base_workdir


    @SYNC_CALL
    def init_instance (self, adaptor_state, type) :

        if not type.lower () in (schema.lower() for schema in _ADAPTOR_SCHEMAS) :
            raise BadParameter \
                    ("the MyProxy context adaptor only handles MyProxy contexts - duh!")

        self.get_api ().type = type

        return self.get_api ()


    @SYNC_CALL
    def _initialize (self, session) :

        # make sure we have server, username, password
        api = self.get_api ()


        # set up the myproxy command
        cmd = ""

        if api.user_pass :
            cmd = "echo %s | myproxy-logon --stdin_pass"  %  api.user_pass
        else :
            cmd = "myproxy-logon --stdin_pass"

        if api.server :
            if ':' in api.server:
                (server, port) = api.server.split (':', 2)
            else:
                server = api.server
                port = "7512"
            if server    : cmd += " --pshost %s"          %  server
            if port      : cmd += " --psport %s"          %  port

        if  api.user_id :
            cmd += " --username %s"        %  api.user_id

        if  api.life_time and api.life_time > 0 :
            cmd += " --proxy_lifetime %s"  %  api.life_time

        # store the proxy in a private location
        proxy_store    = "%s" % self.base_workdir
        proxy_location = "%s/myproxy_%d.x509"  %  (proxy_store, id(self))

        if not os.path.exists (proxy_store) :
            try :
                os.makedirs (proxy_store)
            except OSError as e :
                raise NoSuccess ("could not create myproxy store") from e

        cmd += " --out %s"  %  proxy_location

        expected_result  = "A credential has been received for user %s in %s.\n" \
                         %  (api.user_id, proxy_location)

        process          = subprocess.Popen (cmd, shell=True,
                                             stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE)
        (stdout, stderr) = process.communicate ()

        if expected_result == stdout :
            self._logger.info (stdout)
        else :
            self._logger.info (stderr)
            raise BadParameter ("could not evaluate myproxy context: %s"  %  stderr)

        new_ctx = api.Context ('X509')

        new_ctx.user_proxy = proxy_location
        new_ctx.life_time  = api.life_time

        session.add_context (new_ctx)








__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Provides the SAGA runtime. """

import radical.utils as ru

from ..        import exceptions as rse
from ..version import *


# ------------------------------------------------------------------------------
#
# we set the default of the pty share mode to 'no' on CentOS, as that seems to
# consistently come with old ssh versions which can't handle sharing for sftp
# channels.
#
_share_mode_default = 'auto'
try:
    import subprocess as sp
    _p = sp.Popen ('lsb_release -a | grep "Distributor ID" | cut -f 2 -d ":"',
                   stdout=sp.PIPE, stderr=sp.STDOUT, shell=True)
    _os_flavor = _p.communicate()[0].strip().lower()

    if 'centos'  in _os_flavor or \
       'cent_os' in _os_flavor or \
       'cent-os' in _os_flavor or \
       'cent os' in _os_flavor :
        _share_mode_default = 'no'

except Exception:
    # we ignore this then -- we are relatively sure that the above should work
    # on CentOS...
    pass


# ------------------------------------------------------------------------------
#
class Engine(object, metaclass=ru.Singleton):
    """
    Represents the SAGA engine runtime system.

    The Engine is a singleton class that takes care of adaptor
    loading and management, and which binds adaptor instances to
    API object instances.   The Engine singleton is implicitly
    instantiated as soon as SAGA is imported into Python.  It
    will, on creation, load all available adaptors.  Adaptors
    modules MUST provide an 'Adaptor' class, which will register
    the adaptor in the engine with information like these
    (simplified)::

      _ADAPTOR_INFO = {
        'name'    : _adaptor_name,
        'version' : 'v1.3'
        'schemas' : ['fork', 'local']
        'cpis'    : [{
          'type'    : 'radical.saga.job.Job',
          'class'   : 'LocalJob',
          },
          {
          'type'    : 'radical.saga.job.Service',
          'class'   : 'LocalJobService',
          }
        ]
      }

    where 'class' points to the actual adaptor classes, and
    'schemas' lists the URL schemas for which those adaptor
    classes should be considered.  Note that schemas are case
    insensitive.  More details on the adaptor registration process
    and on adaptor meta data can be found in the adaptors writer
    guide.

    :todo: add link to adaptor writers documentation.

    While loading adaptors, the Engine builds up an internal
    registry of adaptor classes, hierarchically sorted like this
    (simplified)::

      _adaptor_registry =
      {
          'job' :
          {
              'gram' : [<gram job  adaptor, gram job adaptor class>]
              'ssh'  : [<ssh  job  adaptor, ssh  job adaptor class>]
              'http' : [<aws  job  adaptor, aws  job adaptor class>,
                        <occi job  adaptor, occi job adaptor class>]
              ...
          },
          'file' :
          {
              'ftp'  : <ftp file adaptor, ftp file adaptor class>
              'scp'  : <scp file adaptor, scp file adaptor class>
              ...
          },
          ...
      }

    to enable simple lookup operations when binding an API object
    to an adaptor class instance.  For example, a
    'saga.job.Service('http://remote.host.net/')' constructor
    would use (simplified)::

      def __init__ (self, url="", session=None) :

          for (adaptor, adaptor_class) in \
                  self._engine._adaptor_registry{'job'}{url.scheme}

              try :
                  self._adaptor = adaptor_class(self, url, session}

              except saga.Exception e :
                  # adaptor bailed out
                  continue

              else :
                  # successfully bound to adaptor
                  return
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        # Engine manages cpis from adaptors
        self._adaptor_registry = dict()

        # get angine, adaptor and pty configs
        self._cfg      = ru.Config('radical.saga.engine')
        self._pty_cfg  = ru.Config('radical.saga.pty')
        self._registry = ru.Config('radical.saga.registry')

        # Initialize the logging, and log version (this is a singleton!)
        self._logger = ru.Logger('radical.saga')
        self._logger.info('radical.saga         version: %s' % version_detail)

        # load adaptors
        self._load_adaptors()


    # --------------------------------------------------------------------------
    #
    def _load_adaptors (self, inject_registry=None):
        """
        Try to load all adaptors that are registered in saga.engine.registry.py.
        This method is called from the constructor.  As Engine is a singleton,
        this method is called once after the module is first loaded in any
        python application.

        :param inject_registry: Inject a fake registry. *For unit tests only*.
        """


        self._logger.debug ("listing  adaptor registry: %s" % self._registry)

        # check if some unit test wants to use a special registry.  If
        # so, we reset cpi infos from the earlier singleton creation.
        if inject_registry is not None:
            self._adaptor_registry = dict()
            self._registry = {'adaptor_registry' : inject_registry}

        # attempt to load all registered modules
        for module_name in self._registry.get('adaptor_registry', []):

            self._logger.info ("loading  adaptor %s" % module_name)


            # first, import the module
            adaptor_module = None
            try :
                adaptor_module = ru.import_module(module_name)

            except Exception as e:
                self._logger.warning("skip adaptor %s: import failed (%s)",
                                     module_name, e, exc_info=True)
                continue

            # we expect the module to have an 'Adaptor' class
            # implemented, which, on calling 'register()', returns
            # a info dict for all implemented adaptor classes.
            adaptor_instance = None
            adaptor_info     = None

            try:
                adaptor_instance = adaptor_module.Adaptor ()
                adaptor_info     = adaptor_instance.register ()

            except rse.SagaException:
                self._logger.warning("skip adaptor %s: failed to load",
                                     module_name, exc_info=True)
                continue

            except Exception:
                self._logger.warning("skip adaptor %s: init failed",
                                     module_name, exc_info=True)
                continue


            # the adaptor must also provide a sanity_check() method, which sould
            # be used to confirm that the adaptor can function properly in the
            # current runtime environment (e.g., that all pre-requisites and
            # system dependencies are met).
            try:
                adaptor_instance.sanity_check ()

            except Exception:
                self._logger.warning("skip adaptor %s: test failed",
                                     module_name, exc_info=True)
                continue


            # check if we have a valid adaptor_info
            if adaptor_info is None :
                self._logger.warning("skip adaptor %s: invalid adaptor data",
                                     module_name)
                continue


            if  'name'    not in adaptor_info or \
                'cpis'    not in adaptor_info or \
                'version' not in adaptor_info or \
                'schemas' not in adaptor_info    :
                self._logger.warning("skip adaptor %s: incomplete data",
                                     module_name)
                continue


            adaptor_name    = adaptor_info['name']
            adaptor_version = adaptor_info['version']
            adaptor_schemas = adaptor_info['schemas']
            adaptor_enabled = True  # default

            # disable adaptors in 'alpha' or 'beta' versions -- unless
            # the 'load_beta_adaptors' config option is set to True
            if not self._cfg.load_beta_adaptors:

                if 'alpha' in adaptor_version.lower() or \
                   'beta'  in adaptor_version.lower()    :

                    self._logger.warning("skip beta adaptor %s (version %s)",
                                         module_name, adaptor_version)
                    continue


            # get the 'enabled' option in the adaptor's config
            # section (radical.saga.cpi.base) ensures that the option exists,
            # if it is initialized correctly in the adaptor class.
            adaptor_config  = None
            adaptor_enabled = False

            try :
                adaptor_config  = ru.Config('radical.saga.adaptors',
                                            name=adaptor_name)
                adaptor_enabled = adaptor_config.get('enabled', True)

            except rse.SagaException:
                self._logger.warning("skip adaptor %s: init failed",
                                     module_name, exc_info=True)
                continue

            except Exception as e:
                self._logger.warning("skip adaptor %s: init error",
                                     module_name, exc_info=True)
                continue


            # only load adaptor if it is not disabled via config files
            if not adaptor_enabled:
                self._logger.warning("skip adaptor %s: disabled", module_name)
                continue


            # check if the adaptor has anything to register
            if 0 == len (adaptor_info['cpis']) :
                self._logger.warning("skip adaptor %s: adaptor has no cpis",
                                     module_name)
                continue


            # we got an enabled adaptor with valid info - yay!  We can
            # now register all adaptor classes (cpi implementations).
            for cpi_info in adaptor_info['cpis'] :

                # check cpi information details for completeness
                if  'type'  not in cpi_info or \
                    'class' not in cpi_info    :
                    self._logger.warning("skip %s cpi: incomplete info detail",
                                         module_name)
                    continue


                # adaptor classes are registered for specific API types.
                cpi_type  = cpi_info['type']
                cpi_cname = cpi_info['class']
                cpi_class = None

                try :
                    cpi_class = getattr (adaptor_module, cpi_cname)

                except Exception:
                    # this exception likely means that the adaptor does not call
                    # the radical.saga.adaptors.Base initializer (correctly)
                    self._logger.warning("skip adaptor %s: invalid %s",
                                         module_name, cpi_info['class'],
                                         exc_info=True)
                    continue

                # make sure the cpi class is a valid cpi for the given type.
                # We walk through the list of known modules, and try to find
                # a modules which could have that class.  We do the following
                # tests:
                #
                #   cpi_class: ShellJobService
                #   cpi_type:  radical.saga.job.Service
                #   modules:   radical.saga.adaptors.cpi.job
                #   modules:   radical.saga.adaptors.cpi.job.service
                #   classes:   radical.saga.adaptors.cpi.job.Service
                #   classes:   radical.saga.adaptors.cpi.job.service.Service
                #
                #   cpi_class: X509Context
                #   cpi_type:  radical.saga.Context
                #   modules:   radical.saga.adaptors.cpi.context
                #   classes:   radical.saga.adaptors.cpi.context.Context
                #
                # So, we add a 'adaptors.cpi' after the 'saga' namespace
                # element, then append the rest of the given namespace.  If that
                # gives a module which has the requested class, fine -- if not,
                # we add a lower cased version of the class name as last
                # namespace element, and check again.

                # ->   radical .  saga .  job .  Service
                # <- ['radical', 'saga', 'job', 'Service']
                cpi_type_nselems = cpi_type.split ('.')

                if  len(cpi_type_nselems) < 3 or \
                    len(cpi_type_nselems) > 4    :
                    self._logger.warning("skip adaptor %s invalid cpi %s",
                                         module_name, cpi_type)
                    continue

                if  cpi_type_nselems[0] != 'radical' and \
                    cpi_type_nselems[1] != 'saga'    :
                    self._logger.warning("skip adaptor %s: invalid cpi ns %s",
                                         module_name, cpi_type, exc_info=True)
                    continue

                # -> ['radical', 'saga',                    'job', 'Service']
                # <- ['radical', 'saga', 'adaptors', 'cpi', 'job', 'Service']
                cpi_type_nselems.insert (2, 'adaptors')
                cpi_type_nselems.insert (3, 'cpi')

             #  # -> ['radical', 'saga', 'adaptors', 'cpi', 'job',  'Service']
             #  # <- ['radical', 'saga', 'adaptors', 'cpi', 'job'], 'Service'
             #  cpi_type_cname = cpi_type_nselems.pop ()
             #
             #  # -> ['radical', 'saga', 'adaptors', 'cpi', 'job'], 'Service'
             #  # <-  'radical.saga.adaptors.cpi.job
             #  # <-  'radical.saga.adaptors.cpi.job.service
             #  cpi_type_modname_1 = '.'.join (cpi_type_nselems)
             #  cpi_type_modname_2 = '.'.join (cpi_type_nselems + \
             #                                 [cpi_type_cname.lower()])
             #
             #  # does either module exist?
             #  cpi_type_modname = None
             #
             #  if  cpi_type_modname_1 in sys.modules :
             #      cpi_type_modname = cpi_type_modname_1
             #
             #  if  cpi_type_modname_2 in sys.modules :
             #      cpi_type_modname = cpi_type_modname_2
             #
             #  if  not cpi_type_modname :
             #      self._logger.warning("skip adaptor %s: unknown cpi %s",
             #                           module_name, cpi_type, exc_info=True)
             #      sys.exit()
             #      continue
             #
             #  # so, make sure the given cpi is actually
             #  # implemented by the adaptor class
             #  cpi_ok = False
             #  for name, cpi_obj \
             #      in inspect.getmembers (sys.modules[cpi_type_modname]):
             #      if  name == cpi_type_cname      and \
             #          inspect.isclass (cpi_obj)       :
             #          if  issubclass (cpi_class, cpi_obj) :
             #              cpi_ok = True
             #
             #  if not cpi_ok :
             #      self._logger.warning("skip adaptor %s: no cpi %s (%s)",
             #                           module_name, cpi_class, cpi_type,
             #                            exc_info=True)
             #      continue


                # finally, register the cpi for all its schemas!
                registered_schemas = list()
                for adaptor_schema in adaptor_schemas:

                    adaptor_schema = adaptor_schema.lower ()

                    # make sure we can register that cpi type
                    if cpi_type not in self._adaptor_registry:
                        self._adaptor_registry[cpi_type] = dict()

                    # make sure we can register that schema
                    if adaptor_schema not in self._adaptor_registry[cpi_type]:
                        self._adaptor_registry[cpi_type][adaptor_schema] = []

                    # we register the cpi class, so that we can create
                    # instances as needed, and the adaptor instance,
                    # as that is passed to the cpi class c'tor later
                    # on (the adaptor instance is used to share state
                    # between cpi instances, amongst others)
                    info = {'cpi_cname'        : cpi_cname,
                            'cpi_class'        : cpi_class,
                            'adaptor_name'     : adaptor_name,
                            'adaptor_instance' : adaptor_instance}

                    # make sure this tuple was not registered, yet
                    if info in self._adaptor_registry[cpi_type][adaptor_schema]:
                        self._logger.warning("skip adaptor %s: exists %s: %s",
                                             module_name, cpi_class,
                                             adaptor_instance, exc_info=True)
                        continue

                    self._adaptor_registry[cpi_type] \
                                          [adaptor_schema].append(info)
                    registered_schemas.append(str("%s://" % adaptor_schema))

                self._logger.info("Register adaptor %s for %s API: %s" %
                                 (module_name, cpi_type, registered_schemas))



    # --------------------------------------------------------------------------
    #
    def find_adaptors (self, ctype, schema) :
        ''' Look for a suitable cpi class serving a particular schema

            This method will sift through our adaptor registry (see
            '_load_adaptors()', and dig for any adaptor which marches the given
            api class type and schema.  All matching adaptors are returned (by
            name)
        '''

        if ctype not in self._adaptor_registry :
            return []

        if not schema.lower () in self._adaptor_registry[ctype] :
            return []

        adaptor_names = []
        for info in self._adaptor_registry[ctype][schema.lower ()] :
            adaptor_names.append (info['adaptor_name'])

        return adaptor_names



    # --------------------------------------------------------------------------
    #
    def get_adaptor (self, adaptor_name) :
        ''' Return the adaptor module's ``Adaptor`` class for the given adaptor
            name.

            This method is used if adaptor or API object implementation need to
            interact with other adaptors.
        '''

        for ctype in list(self._adaptor_registry.keys ()) :
            for schema in list(self._adaptor_registry[ctype].keys ()) :
                for info in self._adaptor_registry[ctype][schema] :
                    if info['adaptor_name'] == adaptor_name:
                        return info['adaptor_instance']

        error_msg = "No adaptor named '%s' found" % adaptor_name
        self._logger.error(error_msg)
        raise rse.NoSuccess(error_msg)


    # --------------------------------------------------------------------------
    #
    def bind_adaptor (self, api_instance, ctype, schema,
                      preferred_adaptor, *args, **kwargs) :
        '''
        Look for a suitable adaptor class to bind to, instantiate it, and
        initialize it.

        If 'preferred_adaptor' is not 'None', only that given adaptors is
        considered, and adaptor classes are only created from that specific
        adaptor.
        '''

        if ctype not in self._adaptor_registry:
            error_msg = "No adaptor found for '%s' and URL scheme %s://" \
                                  % (ctype, schema)
            self._logger.error(error_msg)
            raise rse.NotImplemented(error_msg)

        if schema not in self._adaptor_registry[ctype]:
            error_msg = "No adaptor found for '%s' and URL scheme %s://" \
                                  % (ctype, schema)
            self._logger.error(error_msg)
            raise rse.NotImplemented(error_msg)


        # cycle through all applicable adaptors, and try to instantiate
        # a matching one.
        exception = rse.NoSuccess ("binding adaptor failed", api_instance)
        for info in self._adaptor_registry[ctype][schema] :

            cpi_cname        = info['cpi_cname']
            cpi_class        = info['cpi_class']
            adaptor_name     = info['adaptor_name']
            adaptor_instance = info['adaptor_instance']

            try :

                # is this adaptor acceptable?
                if  preferred_adaptor is not None     and \
                    preferred_adaptor != adaptor_instance :

                    # ignore this adaptor
                    self._logger.debug ("bind_adaptor for %s : %s != %s ignored"
                             % (cpi_cname, preferred_adaptor, adaptor_instance))
                    continue


                # instantiate cpi
                cpi_instance = cpi_class (api_instance, adaptor_instance)

              # self._logger.debug("Successfully bound %s.%s to %s" \
              #                  % (adaptor_name, cpi_cname, api_instance))
                return cpi_instance


            except rse.SagaException as e :
                # adaptor class initialization failed - try next one
                exception._add_exception (e)
                self._logger.info("adaptor ctor failed : %s.%s: %s"
                                 % (adaptor_name, cpi_class, str(e)))
                continue
            except Exception as e :
                exception._add_exception (rse.NoSuccess (str(e), api_instance))
                self._logger.info("adaptor ctor failed : %s.%s: %s"
                                 % (adaptor_name, cpi_class, str(e)))
                continue


        self._logger.error("No adaptor found for '%s' and URL scheme '%s'"
                          % (ctype, schema))
        self._logger.info  ("%s" %  (str(exception)))
        raise exception._get_exception_stack ()


    # -----------------------------------------------------------------
    #
    def loaded_adaptors (self):
        return self._adaptor_registry


    # -----------------------------------------------------------------
    #
    def _dump(self):

        import pprint

        print('adaptors')
        pprint.pprint (self._adaptor_registry)








__author__    = "Hangi,Kim"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" IBM LoadLeveler job adaptor implementation
    reference for pbs job adaptor & sge job adaptor implementation
    Hangi, Kim hgkim@kisti.re.kr
"""


import os
import re
import time

from   urllib.parse import parse_qs
from   datetime import datetime

import radical.utils as ru

from .. import base
from .. import cpi

from ...job           import constants  as c
from ...exceptions    import *
from ...              import job        as sj
from ...utils         import pty_shell  as sups
from ..sge.sgejob     import SgeKeyValueParser

SYNC_CALL  = cpi.decorators.SYNC_CALL
ASYNC_CALL = cpi.decorators.ASYNC_CALL


# --------------------------------------------------------------------
#
def _ll_to_saga_jobstate(lljs):
    """ translates a loadleveler one-letter state to saga
        pbs_loadl_comparison.xlsx
    """
    if   lljs == 'C' : return c.DONE
    elif lljs == 'S' : return c.PENDING
    elif lljs == 'ST': return c.PENDING
    elif lljs == 'I' : return c.PENDING
    elif lljs == 'R' : return c.RUNNING
    else             : return c.UNKNOWN



def getId(out):

    t = out.split('\n')

    jobId = None

    for line in t:
        if line.startswith('Job'):
            tmpStr = line.split(' ')
            jobId  = tmpStr[1]
            break

        elif re.search('The job ".+" has been submitted.', line):
            # Format: llsubmit: The job "srv03-ib.443336" has been submitted.
            jobId = re.findall(r'"(.*?)"', line)[0]
            break

    if not jobId:
        raise Exception("Failed to detect jobId.")

    return jobId


# --------------------------------------------------------------------
# some private defs
#
_PTY_TIMEOUT = 2.0

# --------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.loadljob"
_ADAPTOR_SCHEMAS       = ["loadl", "loadl+ssh", "loadl+gsissh"]

# --------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES = {
    "jdes_attributes":   [c.NAME,
                          c.EXECUTABLE,
                          c.ARGUMENTS,
                          c.ENVIRONMENT,
                          c.INPUT,
                          c.OUTPUT,
                          c.ERROR,
                          c.QUEUE,
                          c.PROJECT,
                          c.JOB_CONTACT,
                          c.WALL_TIME_LIMIT,
                          c.WORKING_DIRECTORY,
                          c.TOTAL_PHYSICAL_MEMORY,
                          c.PROCESSES_PER_HOST,
                          c.CANDIDATE_HOSTS,
                          c.TOTAL_CPU_COUNT],
    "job_attributes":    [c.EXIT_CODE,
                          c.EXECUTION_HOSTS,
                          c.CREATED,
                          c.STARTED,
                          c.FINISHED],
    "metrics":           [c.STATE],
    "contexts":          {"ssh": "SSH public/private keypair",
                          "x509": "GSISSH X509 proxy context",
                          "userpass": "username/password pair (ssh)"}
}

# --------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC = {
    "name":          _ADAPTOR_NAME,
    "capabilities":  _ADAPTOR_CAPABILITIES,
    "description":  """
The LoadLeveler adaptor allows to run and manage jobs on ` IBM LoadLeveler<http://www-03.ibm.com/systems/software/loadleveler/>`_
controlled HPC clusters.
""",
    "example": "examples/jobs/loadljob.py",
    "schemas": {"loadl":        "connect to a local cluster",
                "loadl+ssh":    "conenct to a remote cluster via SSH",
                "loadl+gsissh": "connect to a remote cluster via GSISSH"}
}

# --------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA
#
_ADAPTOR_INFO = {
    "name"   : _ADAPTOR_NAME,
    "version": "v0.1",
    "schemas": _ADAPTOR_SCHEMAS,
    "cpis"   : [
                   {
                       "type": "radical.saga.job.Service",
                       "class": "LOADLJobService"
                   },
                   {
                       "type": "radical.saga.job.Job",
                       "class": "LOADLJob"
                   }
               ]
}


###############################################################################
#
# The adaptor class
#
class Adaptor (base.Base):
    """ this is the actual adaptor class, which gets loaded by SAGA (i.e. by
        the SAGA engine), and which registers the CPI implementation classes
        which provide the adaptor's functionality.
    """

    # ----------------------------------------------------------------
    #
    def __init__(self):

        base.Base.__init__(self, _ADAPTOR_INFO)

        self.id_re = re.compile('^\[(.*)\]-\[(.*?)\]$')
        self.epoch = datetime(1970,1,1)

        self.purge_on_start   = self._cfg['purge_on_start']
        self.purge_older_than = self._cfg['purge_older_than']


    # ----------------------------------------------------------------
    #
    def sanity_check(self):
        # FIXME: also check for gsissh
        pass

    # ----------------------------------------------------------------
    #
    def parse_id(self, id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(id)

        if not match or len(match.groups()) != 2:
            raise BadParameter("Cannot parse job id '%s'" % id)

        return (match.group(1), match.group(2))


###############################################################################
#
class LOADLJobService (cpi.job.Service):
    """ implements cpi.job.Service
    """

    # ----------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        self._cpi_base = super(LOADLJobService, self)
        self._cpi_base.__init__(api, adaptor)

        self._adaptor = adaptor

    # ----------------------------------------------------------------
    #
    def __del__(self):

        self.finalize(kill_shell=True)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        """ service instance constructor
        """
        self.rm      = rm_url
        self.session = session
        self.ppn     = 0  # check for remove
        self.jobs    = dict()
        self.cluster_option = ''
        self.energy_policy_tag = None
        self.island_count = None
        self.node_usage = None
        self.network_mpi = None
        self.blocking = None
        self.job_type = 'MPICH'  # TODO: Is this a sane default?
        self.enforce_resource_submission = False
        self.enforce_consumable_cpus = False
        self.enforce_consumable_memory = False
        self.enforce_consumable_virtual_memory = False
        self.enforce_consumable_large_page_memory = False
        self.temp_path = "$HOME/.radical/saga/adaptors/loadl_job"

        # LoadLeveler has two ways of specifying the executable and arguments.
        # - Explicit: the executable and arguments are specified as parameters.
        # - Implicit: the (remainder of the) job script is the task.
        #
        # Currently we don't know how this policy can be detected at runtime.
        # We know that providing both will not work in all cases.
        #
        # As the IBM Red Book documents the explicit exec only,
        # we should use that as a default.
        # Currently we just use a hack to workaround Joule.
        #
        # Note: does this now simply become a Joule hack?
        #
        # TODO: Split script into submission file and script and use that for
        #       explicit exec?
        self.explicit_exec = False

        rm_scheme = rm_url.scheme
        pty_url   = ru.Url (rm_url)

        # this adaptor supports options that can be passed via the
        # 'query' component of the job service URL.
        if rm_url.query is not None:
            for key, val in parse_qs(rm_url.query).items():
                if key == 'cluster':
                    self.cluster_option = " -X %s" % val[0]
                elif key == 'energy_policy_tag':
                    self.energy_policy_tag = val[0]
                elif key == 'island_count':
                    self.island_count = val[0]
                elif key == 'node_usage':
                    self.node_usage = val[0]
                elif key == 'network_mpi':
                    self.network_mpi = val[0]
                elif key == 'blocking':
                    self.blocking = val[0]
                elif key == 'job_type':
                    self.job_type = val[0]
                elif key == 'enforce_consumable_cpus':
                    self.enforce_consumable_cpus = True
                    self.enforce_resource_submission = True
                elif key == 'enforce_consumable_memory':
                    self.enforce_consumable_memory = True
                    self.enforce_resource_submission = True
                elif key == 'enforce_consumable_virtual_memory':
                    self.enforce_consumable_virtual_memory = True
                    self.enforce_resource_submission = True
                elif key == 'enforce_consumable_large_page_memory':
                    self.enforce_consumable_large_page_memory = True
                    self.enforce_resource_submission = True
                elif key == 'explicit_exec':
                    self.explicit_exec = True

        # we need to extract the scheme for PTYShell. That's basically the
        # job.Service Url without the loadl+ part. We use the PTYShell to execute
        # loadleveler commands either locally or via gsissh or ssh.
        if rm_scheme == "loadl":
            pty_url.scheme = "fork"
        elif rm_scheme == "loadl+ssh":
            pty_url.scheme = "ssh"
        elif rm_scheme == "loadl+gsissh":
            pty_url.scheme = "gsissh"

        # these are the commands that we need in order to interact with Load
        # Leveler.  the adaptor will try to find them during initialize(self)
        # and bail out in case they are note avaialbe.
        self._commands = {'llq'     : None,
                          'llsubmit': None,
                          'llcancel': None}

        self.shell = sups.PTYShell(pty_url, self.session)

        # self.shell.set_initialize_hook(self.initialize)
        # self.shell.set_finalize_hook(self.finalize)

        self.initialize()

        return self.get_api ()

    # ----------------------------------------------------------------
    #
    def close (self) :
        if  self.shell :
            self.shell.finalize (True)


    # ----------------------------------------------------------------
    #
    def initialize(self):
        # check if all required loadleveler tools are available
        for cmd in self._commands:
            ret, out, _ = self.shell.run_sync("which %s " % cmd)
            self._logger.info(ret)
            self._logger.info(out)
            if ret != 0:
                raise NoSuccess("Error finding LoadLeveler tools: %s" % out)
            else:
                path = out.strip()  # strip removes newline
                ret, out, _ = self.shell.run_sync("%s -v" % cmd)
                if ret != 0:
                    raise NoSuccess("Error finding LoadLeveler tools: %s" % out)
                else:
                    # version is reported as: "version: x.y.z"
                    version = out.strip().split()[1]

                    # add path and version to the command dictionary
                    self._commands[cmd] = {"path":    path,
                                           "version": version}

        self._logger.info("Found LoadLeveler tools: %s" % self._commands)

        # see if we can get some information about the cluster, e.g.,
        # different queues, number of processes per node, etc.
        # TODO: this is quite a hack. however, it *seems* to work quite
        #       well in practice.
        # modi by hgkim

        # purge temporary files
        if self._adaptor.purge_on_start:
            cmd = "find $HOME/.radical/saga/adaptors/loadl_job" \
                  " -type f -mtime +%d -print -delete | wc -l" \
                % self._adaptor.purge_older_than
            ret, out, _ = self.shell.run_sync(cmd)
            if ret == 0 and out != "0":
                self._logger.info("Purged %s temporary files" % out)

    # ----------------------------------------------------------------
    #
    def finalize(self, kill_shell=False):
        if  kill_shell :
            if  self.shell :
                self.shell.finalize (True)

    def __remote_mkdir(self, path):
        """
        Creates a directory on the remote host.
        :param path: the remote directory to be created.
        """
        # check if the path exists
        ret, out, _ = self.shell.run_sync(
                      "(test -d %s && echo -n 0) || (mkdir -p %s && echo -n 1)"
                    % (path, path))

        if ret == 0 and out == "1":
            self._logger.info("Remote directory created: %s" % path)
        elif ret != 0:
            # something went wrong
            raise NoSuccess("Couldn't create remote directory - %s\n%s" % (out, path))


    def __remote_job_info_path(self, loadl_job_id="$LOADL_JOB_NAME"):
        """
        Returns the path of the remote job info file.
        :param loadl_job_id: the LoadLeveler job id.
        if omitted an environment variable representing the job id will be used.
        :return: path to the remote job info file
        """

        return "%s/%s" % (self.temp_path, loadl_job_id)

    def __clean_remote_job_info(self, loadl_job_id):
        """
        Removes the temporary remote file containing job info.
        :param loadl_job_id: the LoadLeveler job id
        """

        path = self.__remote_job_info_path(loadl_job_id)
        ret, out, _ = self.shell.run_sync("rm %s" % path)
        if ret != 0:
            self._logger.debug("Remote job info couldn't be removed: %s" % path)

    def __get_remote_job_info(self, loadl_job_id):
        """
        Obtains the job info from a temporary remote file created by the
        llsubmit script.
        :param loadl_job_id: the LoadLeveler job id
        :return: a dictionary with the job info
        """
        ret, out, _ = self.shell.run_sync("cat %s"
                    % self.__remote_job_info_path(loadl_job_id))
        if ret != 0:
            return None

        qres = SgeKeyValueParser(out, key_suffix=":").as_dict()

        if   "signal"          in qres   : state = c.CANCELED
        elif "exit_status" not in qres   : state = c.RUNNING
        elif not int(qres["exit_status"]): state = c.DONE
        else                             : state = c.FAILED

        job_info = {
                    'state'       : state,
                    'exec_hosts'  : qres.get("hostname"),
                    'create_time' : qres.get("qsub_time"),
                    'start_time'  : qres.get("start_time"),
                    'end_time'    : qres.get("end_time"),
                    'returncode'  : int(qres.get("exit_status", -1)),
                    'gone'        : False
                   }

        return job_info

    def __generate_llsubmit_script(self, jd):
        """
        generates a IMB LoadLeveler script from a SAGA job description
        :param jd: job descriptor
        :return: the llsubmit script
        """
        loadl_params = ''
        exec_string = ''
        args_strings = ''

        if jd.executable is not None:
            exec_string = "%s" % (jd.executable)
        if jd.arguments is not None:
            for arg in jd.arguments:
                args_strings += "%s " % (arg)

        if jd.name is not None:
            loadl_params += "#@ job_name = %s \n" % jd.name

        if jd.environment is not None:
            variable_list = ''
            for key in list(jd.environment.keys()):
                variable_list += "%s=%s;" % (key, jd.environment[key])
            loadl_params += "#@ environment = %s \n" % variable_list

        # Energy
        if self.energy_policy_tag:
            loadl_params += "#@ energy_policy_tag = %s\n" % self.energy_policy_tag
            loadl_params += "#@ minimize_time_to_solution = yes\n"

        if jd.working_directory is not None:
            loadl_params += "#@ initialdir = %s\n" % jd.working_directory
        if jd.output is not None:
            loadl_params += "#@ output = %s\n" % jd.output
        if jd.error is not None:
            loadl_params += "#@ error = %s\n" % jd.error
        if jd.wall_time_limit is not None:
            hours = int(jd.wall_time_limit / 60)
            minutes = jd.wall_time_limit % 60
            loadl_params += "#@ wall_clock_limit = %s:%s:00\n" \
                % (str(hours), str(minutes))

        if jd.total_cpu_count is None:
            # try to come up with a sensible (?) default value
            jd.total_cpu_count = 1
        else:
            if jd.total_cpu_count > 1:
                if self.job_type not in ['bluegene']:
                    # 'bluegene' and total_tasks dont live well together
                    loadl_params += "#@ total_tasks = %s\n" % jd.total_cpu_count

                loadl_params += "#@ job_type = %s\n" % self.job_type

        if self.job_type == 'bluegene':
            BGQ_CORES_PER_NODE = 16  # Only true for BG/Q
            if jd.total_cpu_count % BGQ_CORES_PER_NODE > 0:
                raise Exception("#cores requested is no multiple of 16.")
            loadl_params += "#@ bg_size = %d\n" \
                          % (jd.total_cpu_count / BGQ_CORES_PER_NODE)

        if self.blocking:
            loadl_params += "#@ blocking = %s\n" % self.blocking

        if self.enforce_resource_submission:

            loadl_params += "#@ resources ="

            if self.enforce_consumable_cpus:
                loadl_params += " ConsumableCpus(%d)" % jd.total_cpu_count

            if self.enforce_consumable_memory:
                if jd.total_physical_memory is None:
                    raise Exception("total_physical_memory is not set, but required by enforce_consumable_memory.")
                loadl_params += " ConsumableMemory(%dmb)" % jd.total_physical_memory

            if self.enforce_consumable_large_page_memory:
                # TODO: Not sure how to get a sensible value for this
                if jd.total_physical_memory is None:
                    raise Exception("total_physical_memory is not set, but required by enforce_consumable_large_page_memory.")
                loadl_params += " ConsumableLargePageMemory(%dmb)" % jd.total_physical_memory

            if self.enforce_consumable_virtual_memory:
                # TODO: Not sure how to get a sensible value for this
                if jd.total_physical_memory is None:
                    raise Exception("total_physical_memory is not set, but required by enforce_consumable_virtual_memory.")
                loadl_params += " ConsumableVirtualMemory(%dmb)" % jd.total_physical_memory

            loadl_params += "\n"

        # Number of islands to allocate resources on, can specify a number, or a min/max
        if self.island_count:
            loadl_params += "#@ island_count = %s\n" % self.island_count

        # Specify network configuration
        if self.network_mpi:
            loadl_params += "#@ network.MPI = %s\n" % self.network_mpi

        # Specify node usage policy
        if self.node_usage:
            loadl_params += "#@ node_usage = %s\n" % self.node_usage

        if jd.job_contact is not None:
            if len(jd.job_contact) > 1:
                raise Exception("Only one notify user supported.")
            loadl_params += "#@ notify_user = %s\n" % jd.job_contact[0]
            loadl_params += "#@ notification = always\n"

        # some default (?) parameter that seem to work fine everywhere...
        if jd.queue is not None:
            loadl_params += "#@ class = %s\n" % jd.queue
        else:
            loadl_params += "#@ class = edison\n"

        # finally, we 'queue' the job
        loadl_params += "#@ queue\n"

        # Job info, executable and arguments
        job_info_path = self.__remote_job_info_path()

        script_body = [
            'function aborted() {',
            '  echo Aborted with signal $1.',
            '  echo "signal: $1" >>%s' % job_info_path,
            '  echo "end_time: $(LC_ALL=en_US.utf8 date \'+%%s\')" >>%s' % job_info_path,
            '  exit -1',
            '}',
            'mkdir -p %s' % self.temp_path,
            'for sig in SIGHUP SIGINT SIGQUIT SIGTERM SIGUSR1 SIGUSR2; do trap "aborted $sig" $sig; done',
            'echo "hostname: $HOSTNAME" > %s' % job_info_path,
            'echo "qsub_time: %s"       >>%s' % (time.time(), job_info_path),
            'echo "start_time: $(LC_ALL=en_US.utf8 date \'+%%s\')" >>%s' % job_info_path
        ]

        script_body += ['%s %s' % (exec_string, args_strings)]

        script_body += [
            'echo "exit_status: $?" >>%s' % job_info_path,
            'echo "end_time: $(LC_ALL=en_US.utf8 date \'+%%s\')" >>%s' % job_info_path
        ]

        # convert exec and args into an string and
        # escape all double quotes and dollar signs, otherwise 'echo |'
        # further down won't work.
        # only escape '$' in args and exe. not in the params
        script_body = "\n".join(script_body).replace('$', '\\$')

        # Dirty Trick for Joule: it expects an "executable" parameter,
        # but doesn't really need it, therefore we pass it after the queue
        # parameter, where it is not used anymore.
        if self.explicit_exec:
            loadl_params += "#@ executable = BOGUS\n"

        loadlscript = "\n%s%s" % (loadl_params, script_body)

        return loadlscript.replace('"', '\\"')

    # ----------------------------------------------------------------
    #
    def _job_run(self, jd):
        """ runs a job via llsubmit
        """

        try:
            # create a LoadLeveler job script from SAGA job description
            script = self.__generate_llsubmit_script(jd)

            self._logger.debug("Generated LoadLeveler script: %s" % script)
        except Exception as e:
            raise BadParameter('error creating batch script') from e

        # try to create the working/output/error directories (if defined)
        # WARNING: this assumes a shared filesystem between login node and
        #           compute nodes.
        if jd.working_directory is not None and len(jd.working_directory) > 0:
            self.__remote_mkdir(jd.working_directory)

        if jd.output is not None and len(jd.output) > 0:
            self.__remote_mkdir(os.path.dirname(jd.output))

        if jd.error is not None and len(jd.error) > 0:
            self.__remote_mkdir(os.path.dirname(jd.error))

        # submit the LoadLeveler script
        # Now we want to execute the script. This process consists of two steps:
        # (1) we create a temporary file with 'mktemp' and write the contents of
        #     the generated Load Leveler script into it
        # (2) we call 'llsubmit <tmpfile>' to submit the script to the queueing system
        cmdline = """SCRIPTFILE=`mktemp -t RS-LOADLJobScript.XXXXXX` &&  echo "%s" > $SCRIPTFILE && %s%s $SCRIPTFILE && rm -f $SCRIPTFILE""" %  (script, self._commands['llsubmit']['path'], self.cluster_option)
        self._logger.info("cmdline: %r", cmdline)
        ret, out, _ = self.shell.run_sync(cmdline)

        if ret != 0:
            # something went wrong
            raise NoSuccess("Error running 'llsubmit' job: %s. Script: %s"
                            % (out, script))

        else:
            # stdout contains the job id
            #job_id = "[%s]-[%s]" % (self.rm, out.strip().split('.')[0])
            job_id = "[%s]-[%s]" % (self.rm, getId(out))
            self._logger.info("Submitted LoadLeveler job with id: %s" % job_id)

            # add job to internal list of known jobs.
            self.jobs[job_id] = {
                'state':        c.PENDING,
                'exec_hosts':   None,
                'returncode':   None,
                'create_time':  None,
                'start_time':   None,
                'end_time':     None,
                'gone':         False
            }

            return job_id

    # ----------------------------------------------------------------
    #
    def _retrieve_job(self, job_id, max_retries=10):
        """ see if we can get some info about a job that we don't
            know anything about
            refactoring by referencing sgejob.py
        """
        rm, pid = self._adaptor.parse_id(job_id)

        # run the LoadLeveler 'llq' command to get some info about our job
        ret, out, _ = self.shell.run_sync("%s -j %s -r %%st %%dd %%cc %%jt %%c %%Xs" %
                                          (self._commands['llq']['path'], pid))
        # output is something like
        # R!03/25/2014 13:47!!Serial!normal!kisti.kim
        # OR
        # llq: There is currently no job status to report.
        if ret != 0:
            raise NoSuccess("Couldn't reconnect to job '%s': %s" % (job_id, out))

        else:
            # the job seems to exist on the backend. let's gather some data
            job_info = {
                'state':        c.UNKNOWN,
                'exec_hosts':   None,
                'returncode':   None,
                'create_time':  None,
                'start_time':   None,
                'end_time':     None,
                'gone':         False
            }

          # lastStr = out.rstrip().split('\n')[-1]
            lastStr = out.rstrip()
            self._logger.debug(lastStr)

            if lastStr.startswith('llq:'):
                # llq: There is currently no job status to report

                job_info = None
                retries  = 0

                while job_info is None and retries < max_retries:

                    job_info = self.__get_remote_job_info(pid)
                  # print("llq:", job_info)

                    if job_info is None and retries > 0:
                        message = "__get_remote_job_info get None, pid: %s and retries: %d" % (pid, retries)
                        self._logger.debug(message)
                        # Exponential back-off
                        time.sleep(2**retries)

                    retries += 1

                if job_info is None:
                    raise NoSuccess("__get_remote_job_info exceed %d times(s), "
                                    "pid: %s" % (max_retries, pid))

                self._logger.info("_retrieve_job: %r", job_info)

            else:
                # job is still in the queue
                results = lastStr.split('!')
                self._logger.info("results: %r",results)

                job_info['state']      = _ll_to_saga_jobstate(results[0])
                job_info['returncode'] = None  # still running
                job_info['start_time'] = results[1]
              # job_info['exec_hosts'] = results[5]

            return job_info

    # ----------------------------------------------------------------
    #
    def _job_get_info(self, job_id):
        """ get job attributes via llq
        """

        # if we don't have the job in our dictionary, we don't want it
        if job_id not in self.jobs:
            raise NoSuccess("Unknown job ID: %s. Can't update state." % job_id)

        # prev. info contains the info collect when _job_get_info
        # was called the last time
        prev_info = self.jobs[job_id]

        # if the 'gone' flag is set, there's no need to query the job
        # state again. it's gone forever
        if prev_info['gone'] is True:
            self._logger.warning("Job information is not available anymore.")
            return prev_info

        # if the job is in a terminal state don't expect it to change anymore
        if prev_info["state"] in [c.CANCELED, c.FAILED, c.DONE]:
            return prev_info

        # retrieve updated job information
        curr_info = self._retrieve_job(job_id)
        if curr_info is None:
            prev_info["gone"] = True
            return prev_info

        # update the job info cache and return it
        self.jobs[job_id] = curr_info
        return curr_info

    # ----------------------------------------------------------------
    #
    def _job_get_state(self, job_id):
        """ get the job's state
        """
        # check if we have already reach a terminal state
        if self.jobs[job_id]['state'] in [c.CANCELED, c.FAILED, c.DONE]:
            return self.jobs[job_id]['state']

        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['state']

    # ----------------------------------------------------------------
    #
    def _job_get_exit_code(self, job_id):
        """ get the job's exit code
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['returncode'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['returncode']

    # ----------------------------------------------------------------
    #
    def _job_get_execution_hosts(self, job_id):
        """ get the job's exit code
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['exec_hosts'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['exec_hosts']

    # ----------------------------------------------------------------
    #
    def _job_get_create_time(self, job_id):
        """ get the job's creation time
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['create_time'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['create_time']

    # ----------------------------------------------------------------
    #
    def _job_get_start_time(self, job_id):
        """ get the job's start time
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['start_time'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['start_time']

    # ----------------------------------------------------------------
    #
    def _job_get_end_time(self, job_id):
        """ get the job's end time
        """
        # check if we can / should update
        if not self.jobs[job_id]['gone'] and not self.jobs[job_id]['end_time']:
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['end_time']

    # ----------------------------------------------------------------
    #
    def _job_cancel(self, job_id):
        """ cancel the job via 'llcancel'
        """
        rm, pid = self._adaptor.parse_id(job_id)

        ret, out, _ = self.shell.run_sync("%s%s %s\n" \
            % (self._commands['llcancel']['path'], self.cluster_option, pid))

        if ret != 0:
            raise NoSuccess("Error canceling job via 'llcancel': %s" % out)

        #self.__clean_remote_job_info(pid)

        # assume the job was succesfully canceld
        self.jobs[job_id]['state'] = c.CANCELED

    # ----------------------------------------------------------------
    #
    def _job_wait(self, job_id, timeout):
        """ wait for the job to finish or fail
        """

        time_start = time.time()
        time_now   = time_start
        rm, pid    = self._adaptor.parse_id(job_id)

        while True:
            state = self._job_get_state(job_id=job_id)

            if state == c.UNKNOWN :
                raise IncorrectState("cannot get job state")

            if state in [c.DONE, c.FAILED, c.CANCELED]:
              # self.__clean_remote_job_info(pid)
                return True
            # avoid busy poll
            time.sleep(0.5)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        """ implements cpi.job.Service.get_url()
        """
        # check that only supported attributes are provided
        for attribute in jd.list_attributes():
            if attribute not in _ADAPTOR_CAPABILITIES["jdes_attributes"]:
                raise BadParameter("'jd.%s' is not supported by this adaptor"
                                   % attribute)

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         "job_description": jd,
                         "job_schema":      self.rm.schema,
                         "reconnect":       False
                         }

        return sj.Job(_adaptor=self._adaptor,
                            _adaptor_state=adaptor_state)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, jobid):
        """ Implements cpi.job.Service.get_job()
        """

        self._logger.info("get_job: %r", jobid)
        # try to get some information about this job and throw it into
        # our job dictionary.
        self.jobs[jobid] = self._retrieve_job(jobid)

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         # TODO: fill job description
                         "job_description": sj.Description(),
                         "job_schema":      self.rm.schema,
                         "reconnect":       True,
                         "reconnect_jobid": jobid
                         }

        return sj.Job(_adaptor=self._adaptor,
                            _adaptor_state=adaptor_state)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        """ implements cpi.job.Service.get_url()
        """
        return self.rm

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        """ implements cpi.job.Service.list()
        """
        ids = []

        ret, out, _ = self.shell.run_sync("unset GREP_OPTIONS; %s | grep `whoami`" %
                                          self._commands['llq']['path'])

        if ret != 0 and len(out) > 0:
            raise NoSuccess("failed to list jobs via 'llq': %s" % out)

        elif ret != 0 and len(out) == 0:
            # llq | grep `` exits with 1 if the list is empty
            pass

        else:
            for line in out.split("\n"):
                # output looks like this:
                # v4c064.8637.0            ydkim       3/27 13:33 R  50  normal       v4c064
                # v4c064.8638.0            ydkim       3/27 13:37 R  50  normal       v4c064
                # v4c064.8639.0            ydkim       3/27 13:37 R  50  normal       v4c065
                # v4c064.8640.0            ydkim       3/27 13:37 R  50  normal       v4c065
                # v4c064.8641.0            ydkim       3/27 13:37 I  50  normal
                lineArray=line.split()
                if len(lineArray) > 1:
                    # lineArray[0] : v4c064.8637.0
                    tmpStr=lineArray[0].split('.')
                    jobid = "[%s]-[%s]" % (self.rm, ".".join(tmpStr[:2]))
                    ids.append(str(jobid))

        return ids


  # # ----------------------------------------------------------------
  # #
  # def container_run (self, jobs) :
  #     self._logger.debug ("container run: %s"  %  str(jobs))
  #     # TODO: this is not optimized yet
  #     for job in jobs:
  #         job.run ()
  #
  #
  # # ----------------------------------------------------------------
  # #
  # def container_wait (self, jobs, mode, timeout) :
  #     self._logger.debug ("container wait: %s"  %  str(jobs))
  #     # TODO: this is not optimized yet
  #     for job in jobs:
  #         job.wait ()
  #
  #
  # # ----------------------------------------------------------------
  # #
  # def container_cancel (self, jobs, timeout) :
  #     self._logger.debug ("container cancel: %s"  %  str(jobs))
  #     raise NoSuccess ("Not Implemented");


###############################################################################
#
class LOADLJob (cpi.job.Job):
    """ implements cpi.job.Job
    """

    def __init__(self, api, adaptor):

        # initialize parent class
        self._cpi_base = super(LOADLJob, self)
        self._cpi_base.__init__(api, adaptor)

    @SYNC_CALL
    def init_instance(self, job_info):
        """ implements cpi.job.Job.init_instance()
        """
        # init_instance is called for every new sj.Job object
        # that is created
        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        if job_info['reconnect'] is True:
            self._id      = job_info['reconnect_jobid']
            self._name    = self.jd.get(c.NAME)
            self._started = True
        else:
            self._id      = None
            self._name    = self.jd.get(c.NAME)
            self._started = False

        return self.get_api()

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        """ implements cpi.job.Job.get_state()
        """
        if self._started is False:
            # jobs that are not started are always in 'NEW' state
            return c.NEW
        else:
            return self.js._job_get_state(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):
        """ implements cpi.job.Job.wait()
        """
        if self._started is False:
            raise IncorrectState("Can't wait for job that hasn't been started")
        else:
            self.js._job_wait(self._id, timeout)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):
        """ implements cpi.job.Job.cancel()
        """
        if self._started is False:
            raise IncorrectState("Can't wait for job that hasn't been started")
        else:
            self.js._job_cancel(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):
        """ implements cpi.job.Job.run()
        """
        self._id = self.js._job_run(self.jd)
        self._started = True

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):
        """ implements cpi.job.Job.get_service_url()
        """
        return self.js.rm

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):
        """ implements cpi.job.Job.get_id()
        """
        return self._id

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name (self):
        """ Implements cpi.job.Job.get_name() """
        return self._name

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):
        """ implements cpi.job.Job.get_exit_code()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_exit_code(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):
        """ implements cpi.job.Job.get_created()
        """
        if self._started is False:
            return None
        else:
            # FIXME: convert to EPOCH
            return self.js._job_get_create_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):
        """ implements cpi.job.Job.get_started()
        """
        if self._started is False:
            return None
        else:
            # FIXME: convert to EPOCH
            return self.js._job_get_start_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):
        """ implements cpi.job.Job.get_finished()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_end_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):
        """ implements cpi.job.Job.get_execution_hosts()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_execution_hosts(self._id)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Task interface
"""

import inspect

import radical.utils.signatures  as rus
import radical.utils             as ru

from  . import base              as sbase
from  . import exceptions        as se
from  . import attributes        as satt

from .  import constants         as c

STATES = [c.UNKNOWN, c.NEW, c.RUNNING, c.DONE, c.FAILED, c.CANCELED]


# ------------------------------------------------------------------------------
#
class Task (sbase.SimpleBase, satt.Attributes) :

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task',
                  'CPIBase',
                  str,
                  dict,
                  rus.one_of (c.SYNC, c.ASYNC, c.TASK))
    @rus.returns (rus.nothing)
    def __init__ (self, _adaptor, _method_type, _method_context, _ttype) :
        """
        This saga.Task constructor is private.

        ``_adaptor`` references the adaptor class instance from which this task
        was created via an asynchronous function.  Note that the API level
        object instance can be inferred via ``_adaptor.get_api ()``.  Further,
        the adaptor will reference an _adaptor._container class, which will be
        considered the target for bulk operations for this task.

        ``_method_type`` specifies the SAGA API method which task is
        representing.  For example, for the following code::

          d = saga.filesystem.Directory ("file:///")
          t = d.copy ('/etc/passwd', '/tmp/passwd.bak', saga.task.ASYNC)

        The resulting task ``t`` would represent the *'copy'* method.  This is
        required to forward :class:`saga.task.Container` calls to the correct
        bulk method, in this case ``container_copy()``.

        ``_method_context`` describes the context in which the task method is
        running.  It is up to the creator of the task to provide that context --
        in general, it will at least include method parameters.

        ``ttype`` determines in what state the constructor will leave the task:
        ``DONE`` for ``ttype=SYNC``, ``RUNNING`` for ``ttype=ASYNC`` and ``NEW``
        for ``ttype=TASK``.

        If the ``_method_context`` has *exactly* three elements, names
        ``_call``, ``args`` and ``kwargs``, then the created task will wrap
        a :class:`ru.Future` with that ``_call (*_args, **kwargs)``.
        """

        self._base = super  (Task, self)
        self._base.__init__ ()

        self._future         = None
        self._ttype          = _ttype
        self._adaptor        = _adaptor
        self._method_type    = _method_type
        self._method_context = _method_context

        # set attribute interface properties
        self._attributes_extensible    (False)
        self._attributes_allow_private (True)
        self._attributes_camelcasing   (True)

        # register properties with the attribute interface
        self._attributes_register  (c.RESULT,    None,      satt.ANY,  satt.SCALAR, satt.READONLY)
        self._attributes_set_getter(c.RESULT,    self.get_result)
        self._attributes_set_setter(c.RESULT,    self._set_result)

        self._attributes_register  (c.EXCEPTION, None,      satt.ANY,  satt.SCALAR, satt.READONLY)
        self._attributes_set_getter(c.EXCEPTION, self.get_exception)
        self._attributes_set_setter(c.EXCEPTION, self._set_exception)

        self._attributes_register  (c.STATE,     c.UNKNOWN, satt.ENUM, satt.SCALAR, satt.READONLY)
        self._attributes_set_enums (c.STATE,     STATES)
        self._attributes_set_getter(c.STATE,     self.get_state)
        self._attributes_set_setter(c.STATE,     self._set_state)

        self._set_state (c.NEW)

        # check if this task is supposed to wrap a callable in a future
        if  '_call'   in self._method_context :

            call   = self._method_context['call']
            args   = self._method_context.get('_args',   list())
            kwargs = self._method_context.get('_kwargs', dict())

            # if the called function expects a task handle, provide it.
            if  '_from_task' in inspect.getargspec (call)[0] and \
                '_from_task' not in kwargs :
                kwargs['_from_task'] = self

            self._future = ru.Future (call=call, args=args, kwargs=kwargs)


        # ensure task goes into the correct state
        if self._ttype == c.SYNC :
            self.run  ()
            self.wait ()
        elif self._ttype == c.ASYNC :
            self.run  ()
        elif self._ttype == c.TASK :
            pass


    # --------------------------------------------------------------------------
    #
    def __eq__ (self, other) :

        # we need to define __eq__ to ensure that tasks and jobs with
        # the same # dict representation can be added to Python containers
        # (see Container.add()).  Note that we cannot use `self.get_id()`
        # (or the attribute version `self.id`): job's have the public ID
        # set to `None` until the adaptor assigns a backend specific ID.

        if isinstance(other, Task):
            return self._id == other._id

        else:
            return super(Task, self).__eq__(other)


    # --------------------------------------------------------------------------
    #
    def get_id(self):

        # implement get_id for __eq__
        return self._id


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task')
    @rus.returns (rus.nothing)
    def run      (self) :

        if  self._future :
            self._future.run ()

        else :
            # FIXME: make sure task_run exists.  Should be part of the CPI!
            self._adaptor.task_run (self)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task',
                  rus.optional (float))
    @rus.returns (bool)
    def wait (self, timeout=None) :

        if timeout is None:
            timeout = -1.0  # FIXME

        if self._future :
            self._future.wait (timeout)  # FIXME: timeout?!
            self._set_state   (self._future.state)

        else :
            # FIXME: make sure task_wait exists.  Should be part of the CPI!
            self._adaptor.task_wait (self, timeout)


    # ----------------------------------------------------------------
    #
    @rus.takes   ('Task',
                  float)
    @rus.returns (rus.nothing)
    def cancel (self) :

        if self._future :
            self._future.cancel ()
            self._set_state (c.CANCELED)

        else :
            # FIXME: make sure task_cancel exists.  Should be part of the CPI!
            self._adaptor.task_cancel (self)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task',
                  rus.one_of (c.UNKNOWN, c.NEW, c.RUNNING, c.DONE,
                              c.FAILED,  c.CANCELED))
    @rus.returns (rus.nothing)
    def _set_state (self, state) :

        if state not in STATES:
            raise se.BadParameter ("invalid task state '%s'" % state)

        self._attributes_i_set (self._attributes_t_underscore (c.STATE),
                                                              state, force=True)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task')
    @rus.returns (rus.one_of (STATES))
    def get_state (self) :

        if self._future :
            self._set_state (self._future.state)

        return self.state


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task',
                  rus.anything)
    @rus.returns (rus.nothing)
    def _set_result (self, result) :

        self._attributes_i_set(self._attributes_t_underscore (c.RESULT), result, force=True)
        self._attributes_i_set(self._attributes_t_underscore (c.STATE),  c.DONE, force=True)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task')
    @rus.returns (rus.anything)
    def get_result (self) :

        if self.state not in c.FINAL:
            self.wait ()

        assert (self.state in c.FINAL)

        if  self.state == c.FAILED :
            self.re_raise ()
            return

        if self.state == c.CANCELED :
            raise se.IncorrectState ("get_result() invalid on cancelled tasks")

        if  self.state == c.DONE :

            if  self._future :
                self._set_result (self._future.result)

            return self.result


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task',
                  str,
                  rus.anything)
    @rus.returns (rus.nothing)
    def _set_metric (self, metric, value) :

        self._attributes_i_set (self._attributes_t_underscore (metric),
                                value, force=True)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task',
                  se.SagaException)
    @rus.returns (rus.nothing)
    def _set_exception (self, e) :
        self._attributes_i_set (self._attributes_t_underscore (c.EXCEPTION),
                                e, force=True)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task')
    @rus.returns (se.SagaException)
    def get_exception (self) :

        if  self._future :
            self._set_exception (self._future.exception)

        return self.exception


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Task')
    @rus.returns (rus.nothing)
    def re_raise (self) :

        if self.exception :
            raise self.exception


# ------------------------------------------------------------------------------
#
class Container (sbase.SimpleBase, satt.Attributes) :

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container')
    @rus.returns (rus.nothing)
    def __init__ (self) :


        self._base = super  (Container, self)
        self._base.__init__ ()

        # set attribute interface properties
        self._attributes_allow_private (True)
        self._attributes_extensible    (False)
        self._attributes_camelcasing   (True)

        # register properties with the attribute interface
        self._attributes_register   (c.SIZE,    0,  satt.INT,  satt.SCALAR, satt.READONLY)
        self._attributes_set_getter (c.SIZE,    self.get_size)

        self._attributes_register   (c.TASKS,   [], satt.ANY,  satt.VECTOR, satt.READONLY)
        self._attributes_set_getter (c.TASKS,   self.get_tasks)

        self._attributes_register   (c.STATES,  [], satt.ENUM, satt.VECTOR, satt.READONLY)
        self._attributes_set_getter (c.STATES,  self.get_states)

        self._attributes_set_enums  (c.STATES,  STATES)

        # cache for created container instances
        self._containers = {}


    # --------------------------------------------------------------------------
    #
    def __str__ (self) :

        ret  = "["
        for task in self.tasks :
            ret += "'%s', "  %  str(task)
        ret += "]"

        return ret


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container',
                  Task)
    @rus.returns (rus.nothing)
    def add      (self, task) :

        if  not isinstance (task, Task) :

            raise se.BadParameter ("Container handles tasks, not %s"
                                % (type(task)))

        if task not in self.tasks :
            self.tasks.append (task)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container',
                  Task)
    @rus.returns (rus.nothing)
    def remove   (self, task) :

        if task in self.tasks :
            self.tasks.delete (task)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container')
    @rus.returns (rus.nothing)
    def run      (self) :

        if not len (self.tasks) :
            # nothing to do
            return None

        buckets = self._get_buckets ()
        futures = []  # futures running container ops

        # handle all container
        for b in buckets['bound'] :

            # handle all methods
            for m in buckets['bound'][b] :

                tasks    = buckets['bound'][b][m]
                m_name   = "container_%s" % m
                m_handle = None

                for (name, handle) in inspect.getmembers (b,
                                                    predicate=inspect.ismethod):
                    if name == m_name :
                        m_handle = handle
                        break

                if not handle :
                    # Hmm, the specified container can't handle the call after
                    # all -- fall back to the unbound handling
                    buckets['unbound'] += tasks

                else :
                    # hand off to the container function, in a separate task
                    futures.append (ru.Future.Run (m_handle, tasks))


        # handle tasks not bound to a container
        for task in buckets['unbound'] :

            futures.append (ru.Future.Run (task.run))


        # wait for all futures to finish
        for future in futures :
            if  future.is_alive () :
                future.join ()

            if  future.state == c.FAILED :
              # print '===='
              # print future.exception
              # print future.traceback
              # print '===='
                raise se.NoSuccess ("future exception: %s" % (future.exception))


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container',
                  rus.one_of   (c.ANY, c.ALL),
                  rus.optional (float))
    @rus.returns (rus.list_of (Task))
    def wait (self, mode=c.ALL, timeout=None) :

        if timeout is None:
            timeout = -1.0  # FIXME

        if not len (self.tasks) :
            # nothing to do
            return None

        if mode == c.ALL: return self._wait_all (timeout)
        else            : return self._wait_any (timeout)



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container',
                  float)
    @rus.returns (rus.list_of (Task))
    def _wait_any (self, timeout) :

        buckets = self._get_buckets ()
        futures = []  # futures running container ops

        # handle all tasks bound to containers
        for b in buckets['bound'] :

            # handle all methods -- all go to the same 'container_wait' though)
            tasks = []
            for m in buckets['bound'][b] :
                tasks += buckets['bound'][b][m]

            futures.append (ru.Future.Run (b.container_wait, tasks, c.ANY,
                                           timeout))


        # handle all tasks not bound to containers
        for task in buckets['unbound'] :

            futures.append (ru.Future.Run (task.wait, timeout))


        # mode == ANY: we need to watch our futures, and whenever one
        # returns, and declare success.  Note that we still need to get the
        # finished task from the 'winner'-future -- we do that via a Queue
        # object.  Note also that looser futures are not canceled, but left
        # running (FIXME: consider sending a signal at least)

        timeout = 0.01  # seconds, heuristic :-/

        for future in futures :
            future.join (timeout)

            if future.state == c.FAILED :
                raise future.exception

            if not future.is_alive() :
                # future indeed finished -- dig return value from this
                # futures queue
                result = future.result

                # ignore other futures, and simply declare success
                return result



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container',
                  float)
    @rus.returns (rus.list_of (Task))
    def _wait_all (self, timeout) :
        # this method should actually be symmetric to _wait_any, and could
        # almost be mapped to it, but the code below is a kind of optimization
        # (does not need futures, thus simpler code).

        buckets = self._get_buckets ()
        ret     = None

        # handle all tasks bound to containers
        for b in buckets['bound'] :

            # handle all methods -- all go to the same 'container_wait' though)
            tasks = []
            for m in buckets['bound'][b] :
                tasks += buckets['bound'][b][m]

            # TODO: this is semantically not correct: timeout is applied
            #       n times...
            b.container_wait (tasks, c.ALL, timeout)
            ret = tasks[0]

        # handle all tasks not bound to containers
        for task in buckets['unbound'] :
            task.wait ()
            ret = task

        # all done - return random task (first from last container, or last
        # unbound task)
        # FIXME: that task should be removed from the task container
        return ret


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container',
                  rus.optional (float))
    @rus.returns (rus.nothing)
    def cancel   (self, timeout=None) :

        if timeout is None:
            timeout = -1.0  # FIXME

        buckets = self._get_buckets ()
        futures = []  # futures running container ops

        # handle all tasks bound to containers
        for b in buckets['bound'] :

            # handle all methods -- all go to same 'container_cancel' though)
            tasks = []
            for m in buckets['bound'][b] :
                tasks += buckets['bound'][b][m]

            futures.append (ru.Future.Run (b.container_cancel, tasks, timeout))


        # handle all tasks not bound to containers
        for task in buckets['unbound'] :

            futures.append (ru.Future.Run (task.cancel, timeout))


        for future in futures :
            future.join ()


    # ----------------------------------------------------------------
    #
    @rus.takes   ('Container')
    @rus.returns (int)
    def get_size (self) :

        return len (self.tasks)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container',
                  'basestring')
    @rus.returns (Task)
    def get_task (self, id) :

        # FIXME: this should not be a search, but a lookup
        if not id:
            raise se.NoSuccess ("Lookup requires non-empty id (not '%s')" % id)

        for t in self.tasks:
            if t.get_id() == id:
                return t

        raise se.NoSuccess ("task '%s' not found in container" % id)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container')
    @rus.returns (rus.list_of (Task))
    def get_tasks (self) :

        return self.tasks


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Container')
    @rus.returns (rus.list_of (rus.one_of (*STATES)))
    def get_states (self) :

        buckets = self._get_buckets ()
        futures = []  # futures running container ops

        # handle all tasks bound to containers
        for b in buckets['bound'] :

            # handle all methods -- all go to same 'container_get_states' though
            tasks = []
            for m in buckets['bound'][b] :
                tasks += buckets['bound'][b][m]

            futures.append (ru.Future.Run (b.container_get_states, tasks))


        # handle all tasks not bound to containers
        for task in buckets['unbound'] :

            futures.append (ru.Future.Run (task.get_state))


        # We still need to get the states from all futures.
        # FIXME: order
        states = list()

        for future in futures :
            future.join ()

            if future.state == c.FAILED :
                raise future.exception

            # FIXME: what about ordering tasks / states?
            res = future.result

            if res is not None :
                states.append(res)

        return states


    # ----------------------------------------------------------------
    #
    @rus.takes   ('Container')
    @rus.returns (dict)
    def _get_buckets (self) :
        # collective container ops: walk through the task list, and sort into
        # buckets of tasks which have (a) the same task._container, or if that
        # is not set, the same class type (for which one container instance is
        # created).  All tasks were neither is available are handled one-by-one

        buckets = {}
        buckets['unbound'] = list()  # no container adaptor for these [tasks]
        buckets['bound']   = dict()  # dict  of container adaptors [tasks]

        for task in self.tasks :

            if  task._adaptor and task._adaptor._container :

                # the task's adaptor has a valid associated container class
                # which can handle the container ops - great!
                b = task._adaptor._container
                m = task._method_type

                if b not in buckets['bound'] :
                    buckets['bound'][b] = {}

                if m not in buckets['bound'][b] :
                    buckets['bound'][b][m] = []

                buckets['bound'][b][m].append (task)

            else :

                # we have no container to handle this task -- so
                # put it into the fallback list
                buckets['unbound'].append (task)

        return buckets


     # FIXME: add get_apiobject

# ------------------------------------------------------------------------------

